"""OSM Loaders."""

from ._base import OSMLoader
from .osm_online_loader import OSMOnlineLoader
from .osm_pbf_loader import OSMPbfLoader
from .osm_tile_loader import OSMTileLoader

__all__ = ["OSMLoader", "OSMOnlineLoader", "OSMPbfLoader", "OSMTileLoader"]



"""Module for testing srai.h3 module functionality."""

from typing import Any, Callable

import geopandas as gpd
import pytest

from srai.h3 import (
    ring_buffer_geometry,
    ring_buffer_h3_regions_gdf,
)
from srai.regionalizers.geocode import geocode_to_region_gdf
from srai.regionalizers.h3_regionalizer import H3Regionalizer
from tests.h3.conftest import _gdf_noop, _gdf_to_geometry_list, _gdf_to_geoseries
from tests.regionalizers.test_h3_regionalizer import H3_RESOLUTION

# TODO: add expected values, now only checks if runs without errors


@pytest.mark.parametrize(
    "geometry_fixture, resolution, distance",
    [
        ("gdf_single_point", 10, 10),
        ("gdf_multipolygon", H3_RESOLUTION, 2),
        ("gdf_polygons", H3_RESOLUTION, 2),
    ],
)  # type: ignore
@pytest.mark.parametrize(
    "geometry_parser_function",
    [_gdf_noop, _gdf_to_geoseries, _gdf_to_geometry_list],
)  # type: ignore
def test_ring_buffer_geometry(
    geometry_fixture: str,
    resolution: int,
    distance: int,
    geometry_parser_function: Callable[[gpd.GeoDataFrame], Any],
    request: pytest.FixtureRequest,
) -> None:
    """Test checks if ring_buffer_geometry function works."""
    geometry = request.getfixturevalue(geometry_fixture)

    parsed_geometry = geometry_parser_function(geometry)
    ring_buffer_geometry(parsed_geometry, h3_resolution=resolution, distance=distance)


def test_ring_buffer_h3_regions_gdf() -> None:
    """Test checks if ring_buffer_h3_regions_gdf function works."""
    gdf_wro = geocode_to_region_gdf("Wrocław, PL")
    regions_gdf = H3Regionalizer(8).transform(gdf_wro)

    ring_buffer_h3_regions_gdf(regions_gdf, distance=10)



"""Tests for H3Regionalizer."""

from contextlib import nullcontext as does_not_raise
from typing import TYPE_CHECKING, Any
from unittest import TestCase

import pytest

from srai.constants import GEOMETRY_COLUMN
from srai.regionalizers import S2Regionalizer

if TYPE_CHECKING:  # pragma: no cover
    import geopandas as gpd

ut = TestCase()
S2_RESOLUTION = 7


@pytest.fixture  # type: ignore
def expected_s2_indexes() -> list[str]:
    """Get expected s2 indexes."""
    return [
        "0555c",
        "0ffe4",
        "0fff4",
        "0fffc",
        "10004",
        "1000c",
        "10014",
        "1001c",
        "1aaa4",
        "05554",
        "1aaac",
    ]


@pytest.mark.parametrize(  # type: ignore
    "gdf_fixture,expected_s2_indexes_fixture,resolution,expectation",
    [
        ("gdf_polygons", "expected_s2_indexes", S2_RESOLUTION, does_not_raise()),
        ("gdf_multipolygon", "expected_s2_indexes", S2_RESOLUTION, does_not_raise()),
        ("gdf_empty", "expected_s2_indexes", S2_RESOLUTION, pytest.raises(AttributeError)),
        ("gdf_polygons", "expected_s2_indexes", -1, pytest.raises(ValueError)),
        ("gdf_polygons", "expected_s2_indexes", 31, pytest.raises(ValueError)),
        ("gdf_no_crs", "expected_s2_indexes", S2_RESOLUTION, pytest.raises(ValueError)),
    ],
)
def test_transform(
    gdf_fixture: str,
    expected_s2_indexes_fixture: str,
    resolution: int,
    expectation: Any,
    request: Any,
) -> None:
    """Test transform of H3Regionalizer."""
    gdf: gpd.GeoDataFrame = request.getfixturevalue(gdf_fixture)
    s2_indexes: list[str] = request.getfixturevalue(expected_s2_indexes_fixture)
    with expectation:
        gdf_s2 = S2Regionalizer(resolution).transform(gdf)

        ut.assertCountEqual(first=gdf_s2.index.to_list(), second=s2_indexes)
        assert GEOMETRY_COLUMN in gdf_s2



"""Conftest for Joiners."""

import geopandas as gpd
import pandas as pd
import pytest
from shapely import geometry

from srai.constants import FEATURES_INDEX, REGIONS_INDEX, WGS84_CRS


@pytest.fixture  # type: ignore
def no_geometry_gdf() -> gpd.GeoDataFrame:
    """Get empty GeoDataFrame."""
    return gpd.GeoDataFrame()


@pytest.fixture  # type: ignore
def empty_gdf() -> gpd.GeoDataFrame:
    """Get GeoDataFrame with no geometry."""
    return gpd.GeoDataFrame(geometry=[])


@pytest.fixture  # type: ignore
def regions_gdf() -> gpd.GeoDataFrame:
    """Get GeoDataFrame with example regions."""
    regions = gpd.GeoDataFrame(
        geometry=[
            geometry.Polygon([(-1, 0), (-1, -1), (0, -1), (0, 0)]),
            geometry.Polygon([(1, 0), (1, 1), (0, 1), (0, 0)]),
            geometry.Polygon([(-2, -1), (-2, -2), (-1, -2), (-1, -1)]),
            geometry.Polygon([(-2, 0.5), (-2, -0.5), (-1, -0.5), (-1, 0.5)]),
        ],
        crs=WGS84_CRS,
    )
    return regions


@pytest.fixture  # type: ignore
def features_gdf() -> gpd.GeoDataFrame:
    """Get GeoDataFrame with example features."""
    features = gpd.GeoDataFrame(
        [1, 2, 3, 4],
        geometry=[
            geometry.Polygon([(-1.5, 0.5), (-1.5, 0), (-0.5, 0), (-0.5, 0.5)]),
            geometry.Polygon([(-1.5, -1.5), (-1.5, -2.5), (-0.5, -2.5), (-0.5, -1.5)]),
            geometry.Point((0, 0)),
            geometry.Point((-0.5, -0.5)),
        ],
        crs=WGS84_CRS,
    )
    return features


@pytest.fixture  # type: ignore
def joint_multiindex() -> pd.MultiIndex:
    """Get MultiIndex for joint GeoDataFrame."""
    return pd.MultiIndex.from_tuples(
        [(0, 2), (0, 3), (1, 2), (0, 0), (3, 0), (2, 1)], names=[REGIONS_INDEX, FEATURES_INDEX]
    )



"""Utility function for geocoding a name to `regions_gdf`."""

from typing import Any, Union

import geopandas as gpd

from srai._optional import import_optional_dependencies
from srai.constants import REGIONS_INDEX


def geocode_to_region_gdf(
    query: Union[str, list[str], dict[str, Any]], by_osmid: bool = False
) -> gpd.GeoDataFrame:
    """
    Geocode a query to the `regions_gdf` unified format.

    This functions is a wrapper around the `ox.geocode_to_gdf`[1] function from the `osmnx` library.
    For parameters description look into the source documentation.

    Args:
        query (Union[str, List[str], Dict[str, Any]]): Query string(s) or structured dict(s)
            to geocode.
        by_osmid (bool, optional): Flag to treat query as an OSM ID lookup rather than text search.
            Defaults to False.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with geocoded regions.

    References:
        1. https://osmnx.readthedocs.io/en/stable/osmnx.html#osmnx.geocoder.geocode_to_gdf

    Examples:
        Download geometry for a city
        >>> from srai.regionalizers import geocode_to_region_gdf
        >>> geocode_to_region_gdf("Wrocław, PL")
                                                          geometry
        region_id
        Wrocław, Lower Silesian Voivodeship, Poland  POLYGON ((...

        Download geometries for multiple cities

        >>> geocode_to_region_gdf(["New York City", "Washington, DC"])
                                                                    geometry
        region_id
        New York, United States                          MULTIPOLYGON (((...
        Washington, District of Columbia, United States  POLYGON ((...

        Use OSM relation IDs to get geometries.

        >>> geocode_to_region_gdf(["R175342", "R5750005"], by_osmid=True)
                                                                 geometry
        region_id
        Greater London, England, United Kingdom             POLYGON ((...
        Sydney, Council of the City of Sydney, New Sout...  POLYGON ((...
    """
    import_optional_dependencies(
        dependency_group="osm",
        modules=["osmnx"],
    )

    import osmnx as ox

    geocoded_gdf = ox.geocode_to_gdf(query=query, by_osmid=by_osmid, which_result=None)
    regions_gdf = (
        geocoded_gdf[["display_name", "geometry"]]
        .rename(columns={"display_name": REGIONS_INDEX})
        .set_index(REGIONS_INDEX)
    )
    return regions_gdf



"""Tests for Hex2Vec model."""

from contextlib import nullcontext as does_not_raise
from typing import Any

import pytest

from srai.embedders.hex2vec.model import Hex2VecModel


@pytest.mark.parametrize(  # type: ignore
    "layer_sizes,expectation",
    [
        ([], pytest.raises(ValueError)),
        ([15], pytest.raises(ValueError)),
        ([15, 10], does_not_raise()),
        ([15, 10, 5], does_not_raise()),
    ],
)
def test_model_raises_with_incorrect_layer_sizes(layer_sizes: list[int], expectation: Any) -> None:
    """Test if Hex2VecModel raises with incorrect layer_sizes."""
    with expectation:
        Hex2VecModel(layer_sizes=layer_sizes)


def test_layers_initialized_correctly() -> None:
    """Test if Hex2VecModel layers are initialized correctly."""
    layer_sizes = [15, 10, 5]

    model = Hex2VecModel(layer_sizes=layer_sizes)
    for i, layer in enumerate(model.encoder):
        if i % 2 == 0:
            assert layer.in_features == layer_sizes[i // 2]
            assert layer.out_features == layer_sizes[i // 2 + 1]



"""Conftest for H3 tests."""

import geopandas as gpd
import pytest
from shapely import geometry
from shapely.geometry.base import BaseGeometry

from srai.constants import GEOMETRY_COLUMN, WGS84_CRS


@pytest.fixture  # type: ignore
def gdf_single_point() -> gpd.GeoDataFrame:
    """Get the point case."""
    return gpd.GeoDataFrame(geometry=[geometry.Point(17.9261, 50.6696)], crs=WGS84_CRS)


@pytest.fixture  # type: ignore
def expected_point_h3_index() -> list[str]:
    """Get expected h3 index for the point case."""
    return [
        "8a1e23c44b5ffff",
    ]


@pytest.fixture  # type: ignore
def gdf_polygons() -> gpd.GeoDataFrame:
    """Get GeoDataFrame with two polygons."""
    return gpd.GeoDataFrame(
        geometry=[
            geometry.Polygon(
                shell=[
                    (-1, 0),
                    (0, 0.5),
                    (1, 0),
                    (1, 1),
                    (0, 1),
                ],
                holes=[
                    [
                        (0.8, 0.9),
                        (0.9, 0.55),
                        (0.8, 0.3),
                        (0.5, 0.4),
                    ]
                ],
            ),
            geometry.Polygon(shell=[(-0.25, 0), (0.25, 0), (0, 0.2)]),
        ],
        crs=WGS84_CRS,
    )


@pytest.fixture  # type: ignore
def gdf_multipolygon() -> gpd.GeoDataFrame:
    """Get GeoDataFrame with multipolygon."""
    return gpd.GeoDataFrame(
        geometry=[
            geometry.MultiPolygon(
                [
                    (
                        [
                            (-1, 0),
                            (0, 0.5),
                            (1, 0),
                            (1, 1),
                            (0, 1),
                        ],
                        (
                            [
                                [
                                    (0.8, 0.9),
                                    (0.9, 0.55),
                                    (0.8, 0.3),
                                    (0.5, 0.4),
                                ]
                            ]
                        ),
                    ),
                    (
                        [(-0.25, 0), (0.25, 0), (0, 0.2)],
                        (),
                    ),
                ]
            )
        ],
        crs=WGS84_CRS,
    )


@pytest.fixture  # type: ignore
def expected_h3_indexes() -> list[str]:
    """Get expected h3 indexes."""
    return [
        "837559fffffffff",
        "83754efffffffff",
        "83754cfffffffff",
        "837541fffffffff",
        "83755dfffffffff",
        "837543fffffffff",
        "83754afffffffff",
    ]


@pytest.fixture  # type: ignore
def expected_unbuffered_h3_indexes() -> list[str]:
    """Get expected h3 index for the unbuffered case."""
    return [
        "83754efffffffff",
    ]


def _gdf_noop(gdf_fixture: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    return gdf_fixture


def _gdf_to_geoseries(gdf_fixture: gpd.GeoDataFrame) -> gpd.GeoSeries:  # noqa: FURB118
    return gdf_fixture[GEOMETRY_COLUMN]


def _gdf_to_geometry_list(gdf_fixture: gpd.GeoDataFrame) -> list[BaseGeometry]:
    return list(gdf_fixture[GEOMETRY_COLUMN])


def _gdf_to_single_geometry(gdf_fixture: gpd.GeoDataFrame) -> BaseGeometry:
    return gdf_fixture[GEOMETRY_COLUMN][0]



"""Tests for folium plotting wrapper."""

from srai.neighbourhoods.h3_neighbourhood import H3Neighbourhood
from srai.plotting.folium_wrapper import plot_neighbours
from srai.regionalizers.geocode import geocode_to_region_gdf
from srai.regionalizers.h3_regionalizer import H3Regionalizer


def test_wroclaw_neighbourhood_edge_case() -> None:
    """Test plotting edge case from H3Neighbourhood example error."""
    gdf_wro = geocode_to_region_gdf("Wrocław, PL")
    regions_gdf = H3Regionalizer(8).transform(gdf_wro)
    neighbourhood_with_regions = H3Neighbourhood(regions_gdf)

    edge_region_id = "881e2050bdfffff"
    neighbours_ids = neighbourhood_with_regions.get_neighbours(edge_region_id)

    plot_neighbours(regions_gdf, edge_region_id, neighbours_ids)



"""Custom exceptions for the SRAI package."""


class SRAIException(Exception):
    """Base class for all SRAI exceptions."""


class ModelNotFitException(SRAIException):
    """Exception raised when a model is not fit."""


class LoadedDataIsEmptyException(SRAIException):
    """Exception when the loaded data returned by the loader is empty."""

    def __init__(self, message: str):  # noqa: D107
        super().__init__()

        self.message = message

    def __str__(self) -> str:  # noqa: D105
        return f"The data returned by the loader is empty. {self.message}"



"""Base class for regionalizers."""

import abc

import geopandas as gpd


class Regionalizer(abc.ABC):
    """Abstract class for regionalizers."""

    @abc.abstractmethod
    def transform(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:  # pragma: no cover
        """
        Regionalize a given GeoDataFrame.

        This one should treat the input as a single region.

        Args:
            gdf (gpd.GeoDataFrame): GeoDataFrame to be regionalized.

        Returns:
            GeoDataFrame with the regionalized data.
        """
        raise NotImplementedError

    def _explode_multipolygons(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Explode multipolygons into multiple polygons.

        Args:
            gdf (gpd.GeoDataFrame): GeoDataFrame to be exploded.

        Returns:
            GeoDataFrame with exploded multipolygons.
        """
        return gdf.explode(index_parts=True).reset_index(drop=True)



"""Geoparquet loader tests."""

from pathlib import Path

import geopandas as gpd
import pytest
from shapely.geometry import box

from srai.constants import GEOMETRY_COLUMN, WGS84_CRS
from srai.loaders import GeoparquetLoader

bbox = box(minx=-180, maxx=180, miny=-90, maxy=90)
bbox_gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [bbox]})


def test_wrong_path_error() -> None:
    """Test checks if cannot load nonexistent file."""
    with pytest.raises(FileNotFoundError):
        GeoparquetLoader().load(
            file_path=Path(__file__).parent / "test_files" / "non_existent.parquet"
        )


def test_correct_path() -> None:
    """Test checks if can load proper file."""
    gdf = GeoparquetLoader().load(
        file_path=Path(__file__).parent / "test_files" / "example.parquet"
    )
    assert len(gdf.index) == 5
    assert gdf.crs.to_epsg() == 4326


def test_wrong_index_value_error() -> None:
    """Test checks if cannot set nonexistent index."""
    with pytest.raises(ValueError):
        GeoparquetLoader().load(
            file_path=Path(__file__).parent / "test_files" / "example.parquet",
            index_column="test_column",
        )


def test_clipped_columns() -> None:
    """Test if returned a subset of columns."""
    gdf = GeoparquetLoader().load(
        file_path=Path(__file__).parent / "test_files" / "example.parquet",
        columns=["continent", "name"],
    )
    assert len(gdf.columns == 3)


def test_setting_index() -> None:
    """Test if properly set the index."""
    gdf = GeoparquetLoader().load(
        file_path=Path(__file__).parent / "test_files" / "example.parquet", index_column="name"
    )
    expected_index = {"Canada", "Fiji", "Tanzania", "United States of America", "W. Sahara"}
    assert expected_index == set(gdf.index)


def test_clipping() -> None:
    """Test if properly clips the data."""
    bbox = box(minx=-106.645646, maxx=-93.508292, miny=25.837377, maxy=36.500704)
    bbox_gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [bbox]}, crs=WGS84_CRS)
    gdf = GeoparquetLoader().load(
        file_path=Path(__file__).parent / "test_files" / "example.parquet", area=bbox_gdf
    )
    assert len(gdf.index) == 1



"""Utility H3 related functions."""

from collections.abc import Iterable
from typing import Literal, Union, overload

import geopandas as gpd
import h3
import numpy as np
import numpy.typing as npt
from h3ronpy import __version__ as h3ronpy_version
from packaging import version
from shapely.geometry import Polygon
from shapely.geometry.base import BaseGeometry

from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS

is_new_h3ronpy_api = version.parse(h3ronpy_version) >= version.parse("0.22.0")

if is_new_h3ronpy_api:
    from h3ronpy import ContainmentMode, cells_to_string, grid_disk
    from h3ronpy.vector import cells_to_wkb_polygons, wkb_to_cells
else:
    from h3ronpy.arrow import cells_to_string, grid_disk
    from h3ronpy.arrow.vector import (
        ContainmentMode,
        cells_to_wkb_polygons,
        wkb_to_cells,
    )

__all__ = [
    "shapely_geometry_to_h3",
    "h3_to_geoseries",
    "h3_to_shapely_geometry",
    "get_local_ij_index",
    "ring_buffer_h3_indexes",
    "ring_buffer_geometry",
    "ring_buffer_h3_regions_gdf",
]


def shapely_geometry_to_h3(
    geometry: Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame],
    h3_resolution: int,
    buffer: bool = True,
) -> list[str]:
    """
    Convert Shapely geometry to H3 indexes.

    Args:
        geometry (Union[BaseGeometry, Iterable[BaseGeometry], GeoSeries, GeoDataFrame]):
            Shapely geometry to be converted.
        h3_resolution (int): H3 resolution of the cells. See [1] for a full comparison.
        buffer (bool, optional): Whether to fully cover the geometries with
            H3 Cells (visible on the borders). Defaults to True.

    Returns:
        List[str]: List of H3 indexes that cover a given geometry.

    Raises:
        ValueError: If resolution is not between 0 and 15.

    References:
        1. https://h3geo.org/docs/core-library/restable/
    """
    if not (0 <= h3_resolution <= 15):
        raise ValueError(f"Resolution {h3_resolution} is not between 0 and 15.")

    wkb = []
    if isinstance(geometry, gpd.GeoSeries):
        wkb = geometry.to_wkb()
    elif isinstance(geometry, gpd.GeoDataFrame):
        wkb = geometry[GEOMETRY_COLUMN].to_wkb()
    elif isinstance(geometry, Iterable):
        wkb = [sub_geometry.wkb for sub_geometry in geometry]
    else:
        wkb = [geometry.wkb]

    containment_mode = ContainmentMode.Covers if buffer else ContainmentMode.ContainsCentroid
    h3_indexes = wkb_to_cells(
        wkb, resolution=h3_resolution, containment_mode=containment_mode, flatten=True
    )

    if is_new_h3ronpy_api:
        h3_indexes = np.unique(h3_indexes.to_numpy())
    else:
        h3_indexes = h3_indexes.unique()

    return [h3.int_to_str(h3_index) for h3_index in h3_indexes.tolist()]


# TODO: write tests (#322)
def h3_to_geoseries(h3_index: Union[int, str, Iterable[Union[int, str]]]) -> gpd.GeoSeries:
    """
    Convert H3 index to GeoPandas GeoSeries.

    Args:
        h3_index (Union[int, str, Iterable[Union[int, str]]]): H3 index (or list of indexes)
            to be converted.

    Returns:
        GeoSeries: Geometries as GeoSeries with default CRS applied.
    """
    if isinstance(h3_index, (str, int)):
        return h3_to_geoseries([h3_index])
    else:
        h3_int_indexes = list(
            h3_cell if isinstance(h3_cell, int) else h3.str_to_int(h3_cell) for h3_cell in h3_index
        )
        return gpd.GeoSeries.from_wkb(cells_to_wkb_polygons(h3_int_indexes), crs=WGS84_CRS)


@overload
def h3_to_shapely_geometry(h3_index: Union[int, str]) -> Polygon: ...


@overload
def h3_to_shapely_geometry(h3_index: Iterable[Union[int, str]]) -> list[Polygon]: ...


# TODO: write tests (#322)
def h3_to_shapely_geometry(
    h3_index: Union[int, str, Iterable[Union[int, str]]],
) -> Union[Polygon, list[Polygon]]:
    """
    Convert H3 index to Shapely polygon.

    Args:
        h3_index (Union[int, str, Iterable[Union[int, str]]]): H3 index (or list of indexes)
            to be converted.

    Returns:
        Union[Polygon, List[Polygon]]: Converted polygon (or list of polygons).
    """
    if isinstance(h3_index, (str, int)):
        coords = h3.cell_to_boundary(h3_index, geo_json=True)
        return Polygon(coords)
    return h3_to_geoseries(h3_index).values.tolist()


@overload
def get_local_ij_index(origin_index: str, h3_index: str) -> tuple[int, int]: ...


@overload
def get_local_ij_index(
    origin_index: str, h3_index: list[str], return_as_numpy: Literal[False]
) -> list[tuple[int, int]]: ...


@overload
def get_local_ij_index(
    origin_index: str, h3_index: list[str], return_as_numpy: Literal[True]
) -> npt.NDArray[np.int8]: ...


# Last fallback needed as per documentation:
# https://mypy.readthedocs.io/en/stable/literal_types.html#literal-types
@overload
def get_local_ij_index(
    origin_index: str, h3_index: list[str], return_as_numpy: bool
) -> Union[list[tuple[int, int]], npt.NDArray[np.int8]]: ...


def get_local_ij_index(
    origin_index: str, h3_index: Union[str, list[str]], return_as_numpy: bool = False
) -> Union[tuple[int, int], list[tuple[int, int]], npt.NDArray[np.int8]]:
    """
    Calculate the local H3 ij index based on provided origin index.

    Wraps H3's cell_to_local_ij function and centers returned coordinates
    around provided origin cell.

    Args:
        origin_index (str): H3 index of the origin region.
        h3_index (Union[str, List[str]]): H3 index of the second region or list of regions.
        return_as_numpy (bool, optional): Flag whether to return calculated indexes as a Numpy array
            or a list of tuples.

    Returns:
        Union[Tuple[int, int], List[Tuple[int, int]], npt.NDArray[np.int8]]: The local ij index of
            the second region (or regions) with respect to the first one.
    """
    origin_coords = h3.cell_to_local_ij(origin_index, origin_index)
    if isinstance(h3_index, str):
        ijs = h3.cell_to_local_ij(origin_index, h3_index)
        return (origin_coords[0] - ijs[0], origin_coords[1] - ijs[1])
    ijs = np.array([h3.cell_to_local_ij(origin_index, h3_cell) for h3_cell in h3_index])
    local_ijs = np.array(origin_coords) - ijs

    if not return_as_numpy:
        local_ijs = [(coords[0], coords[1]) for coords in local_ijs]

    return local_ijs


def ring_buffer_h3_indexes(h3_indexes: Iterable[Union[int, str]], distance: int) -> list[str]:
    """
    Buffer H3 indexes by a given number of k-rings.

    List of provided H3 indexes will be buffered by a given distance.

    Args:
        h3_indexes (Iterable[Union[int, str]]): List of H3 indexes to be buffered.
        distance (int): The k-ring buffer distance in H3 cells.

    Returns:
        List[str]: Buffered list of H3 cells containing both original and new cells.
    """
    assert all(
        h3.is_valid_cell(h3_cell) for h3_cell in h3_indexes
    ), "Not all values in h3_indexes are valid H3 cells."

    h3_int_indexes = list(
        h3_cell if isinstance(h3_cell, int) else h3.str_to_int(h3_cell) for h3_cell in h3_indexes
    )

    if is_new_h3ronpy_api:
        buffered_h3s = np.unique(
            cells_to_string(grid_disk(h3_int_indexes, distance, flatten=True)).to_numpy()
        )
    else:
        buffered_h3s = set(
            cells_to_string(grid_disk(h3_int_indexes, distance, flatten=True)).tolist()
        )

    return list(buffered_h3s)


def ring_buffer_geometry(
    geometry: Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame],
    h3_resolution: int,
    distance: int,
) -> Union[gpd.GeoSeries, BaseGeometry]:
    """
    Buffer a Shapely geometry with H3 cells, and return the bounding geometry.

    If a GeoDataFrame is passed, the geometry column will be used and the return will be a GeoSeries

    Args:
        geometry (Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame]):
            The geometry to buffer.
        h3_resolution (int): The H3 resolution to use.
        distance (int): The k-ring buffer distance in H3 cells.

    Returns:
        Union[gpd.GeoSeries, BaseGeometry]: The buffered geometry.
    """
    if isinstance(geometry, gpd.GeoDataFrame):
        geometry = geometry[GEOMETRY_COLUMN]
        return geometry.apply(lambda x: ring_buffer_geometry(x, h3_resolution, distance))

    if isinstance(geometry, gpd.GeoSeries):
        return geometry.apply(lambda x: ring_buffer_geometry(x, h3_resolution, distance))

    if isinstance(geometry, Iterable):
        return gpd.GeoSeries([ring_buffer_geometry(x, h3_resolution, distance) for x in geometry])

    assert isinstance(geometry, BaseGeometry)
    h3s = shapely_geometry_to_h3(geometry, h3_resolution, buffer=True)
    # buffer all the h3
    buffered_h3s = ring_buffer_h3_indexes(h3s, distance=distance)
    # get the bounding geometry
    return h3_to_geoseries(buffered_h3s).unary_union


def ring_buffer_h3_regions_gdf(regions_gdf: gpd.GeoDataFrame, distance: int) -> gpd.GeoDataFrame:
    """
    Buffer H3 indexes by a given number of k-rings.

    List of provided H3 indexes will be buffered by a given distance.

    Args:
        regions_gdf (gpd.GeoDataFrame): GeoDataFrame with H3 regions from H3Regionalizer.
        distance (int): The k-ring buffer distance in H3 cells.

    Returns:
        gpd.GeoDataFrame: Buffered regions_gdf with new H3 cells added.
    """
    buffered_h3_indexes = ring_buffer_h3_indexes(h3_indexes=regions_gdf.index, distance=distance)
    buffered_gdf_h3 = gpd.GeoDataFrame(
        data={REGIONS_INDEX: buffered_h3_indexes},
        geometry=h3_to_geoseries(buffered_h3_indexes),
        crs=WGS84_CRS,
    ).set_index(REGIONS_INDEX)
    return buffered_gdf_h3



"""Hex2VecEmbedder tests."""

from contextlib import nullcontext as does_not_raise
from pathlib import Path
from typing import Any

import geopandas as gpd
import pytest
from pytest_snapshot.plugin import Snapshot
from pytorch_lightning import seed_everything

from srai.embedders.hex2vec.embedder import Hex2VecEmbedder
from srai.exceptions import ModelNotFitException
from srai.neighbourhoods import H3Neighbourhood
from tests.embedders.hex2vec.constants import ENCODER_SIZES, PREDEFINED_TEST_CASES, TRAINER_KWARGS


@pytest.mark.parametrize(  # type: ignore
    "encoder_sizes,expectation",
    [
        ([150, 75, 50], does_not_raise()),
        ([10, 5], does_not_raise()),
        ([5], does_not_raise()),
        ([], pytest.raises(ValueError)),
        ([-1, 0], pytest.raises(ValueError)),
        ([0], pytest.raises(ValueError)),
    ],
)
def test_checking_encoder_sizes(encoder_sizes: list[int], expectation: Any) -> None:
    """Test that incorrect encoder sizes raise ValueError."""
    with expectation:
        Hex2VecEmbedder(encoder_sizes)


def test_embedder_not_fitted() -> None:
    """Test that Hex2VecEmbedder raises ModelNotFitException if not fitted."""
    embedder = Hex2VecEmbedder()
    with pytest.raises(ModelNotFitException):
        embedder.transform(gpd.GeoDataFrame(), gpd.GeoDataFrame(), gpd.GeoDataFrame())


def test_embedder_default_encoder_sizes() -> None:
    """Test that Hex2VecEmbedder uses default encoder sizes if not specified."""
    embedder = Hex2VecEmbedder()
    assert embedder._encoder_sizes == Hex2VecEmbedder.DEFAULT_ENCODER_SIZES


def test_embedder(snapshot: Snapshot) -> None:
    """Test Hex2VecEmbedder on predefined test cases."""
    test_files_path = Path(__file__).parent / "test_files"
    snapshot.snapshot_dir = test_files_path.as_posix()

    for test_case in PREDEFINED_TEST_CASES:
        name = test_case["test_case_name"]
        seed = test_case["seed"]
        print(name, seed)

        regions_gdf = gpd.read_parquet(test_files_path / f"{name}_regions.parquet")
        features_gdf = gpd.read_parquet(test_files_path / f"{name}_features.parquet")
        joint_gdf = gpd.read_parquet(test_files_path / f"{name}_joint.parquet")
        seed_everything(seed)

        neighbourhood = H3Neighbourhood(regions_gdf)
        embedder = Hex2VecEmbedder(encoder_sizes=ENCODER_SIZES)
        result_df = embedder.fit_transform(
            regions_gdf, features_gdf, joint_gdf, neighbourhood, trainer_kwargs=TRAINER_KWARGS
        )
        result_df.columns = result_df.columns.astype(str)
        print(result_df.head())

        # FIXME(Calychas): readd after making neighbourhoods deterministic.
        # See [#441](https://github.com/kraina-ai/srai/pull/441)
        # snapshot.assert_match(result_df.to_json(
        #     orient="index", indent=True), f"{name}_result.json"
        # )
        assert not result_df.empty
        assert result_df.shape[0] == regions_gdf.shape[0]
        assert result_df.shape[1] == ENCODER_SIZES[-1]



"""
GeoVex Embedder.

This module contains embedder from GeoVex paper[1].

References:
    [1] https://openreview.net/forum?id=7bvWopYY1H
"""

import json
from pathlib import Path
from typing import Any, Optional, TypeVar, Union

import geopandas as gpd
import numpy as np
import pandas as pd

from srai._optional import import_optional_dependencies
from srai.constants import REGIONS_INDEX
from srai.embedders import CountEmbedder, ModelT
from srai.embedders.geovex.dataset import HexagonalDataset
from srai.embedders.geovex.model import GeoVexModel
from srai.exceptions import ModelNotFitException
from srai.loaders.osm_loaders.filters import GroupedOsmTagsFilter, OsmTagsFilter
from srai.neighbourhoods import H3Neighbourhood

T = TypeVar("T")

try:  # pragma: no cover
    from torch.utils.data import DataLoader

except ImportError:
    from srai.embedders._pytorch_stubs import DataLoader


class GeoVexEmbedder(CountEmbedder):
    """GeoVex Embedder."""

    def __init__(
        self,
        target_features: Union[list[str], OsmTagsFilter, GroupedOsmTagsFilter],
        batch_size: Optional[int] = 32,
        neighbourhood_radius: int = 4,
        convolutional_layers: int = 2,
        embedding_size: int = 32,
        convolutional_layer_size: int = 256,
    ) -> None:
        """
        Initialize GeoVex Embedder.

        Args:
            target_features (Union[List[str], OsmTagsFilter, GroupedOsmTagsFilter]): The features
                that are to be used in the embedding. Should be in "flat" format,
                i.e. "<super-tag>_<sub-tag>", or use OsmTagsFilter object.
            batch_size (int, optional): Batch size. Defaults to 32.
            convolutional_layers (int, optional): Number of convolutional layers. Defaults to 2.
            neighbourhood_radius (int, optional): Radius of the neighbourhood. Defaults to 4.
            embedding_size (int, optional): Size of the embedding. Defaults to 32.
            convolutional_layer_size (int, optional): Size of the first convolutional layer.
        """
        import_optional_dependencies(
            dependency_group="torch", modules=["torch", "pytorch_lightning"]
        )

        super().__init__(
            expected_output_features=target_features,
        )

        self._assert_feature_length(self.expected_output_features, convolutional_layer_size)

        self._model: Optional[GeoVexModel] = None
        self._is_fitted = False
        self._r = neighbourhood_radius
        self._embedding_size = embedding_size
        self._convolutional_layers = convolutional_layers
        self._convolutional_layer_size = convolutional_layer_size

        self._batch_size = batch_size

        # save invalid h3s for later
        self._invalid_cells: list[str] = []
        self._dataset: DataLoader = None

    @property
    def invalid_cells(self) -> list[str]:
        """List of invalid h3s."""
        return self._invalid_cells

    def transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:
        """
        Create region embeddings.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Returns:
            pd.DataFrame: Region embeddings.
        """
        self._check_is_fitted()

        neighbourhood = H3Neighbourhood(
            regions_gdf=regions_gdf,
        )

        _, dataloader, self._dataset = self._prepare_dataset(
            regions_gdf,
            features_gdf,
            joint_gdf,
            neighbourhood,
            self._batch_size,
            shuffle=False,
        )

        return self._transform(dataset=self._dataset, dataloader=dataloader)

    def _transform(
        self,
        dataset: HexagonalDataset[T],
        dataloader: Optional[DataLoader] = None,
    ) -> pd.DataFrame:
        if dataloader is None:
            dataloader = DataLoader(dataset, batch_size=self._batch_size, shuffle=False)

        embeddings = [
            self._model.encoder(batch).detach().numpy()  # type: ignore
            for batch in dataloader
        ]
        if dataset.get_invalid_cells():
            print(
                "Warning: Some regions were not able to be encoded, as they don't have"
                f" r={self._r} neighbors."
            )
        df = pd.DataFrame(np.concatenate(embeddings), index=dataset.get_valid_cells())
        df.index.name = REGIONS_INDEX
        return df

    def fit(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
        neighbourhood: H3Neighbourhood,
        learning_rate: float = 0.001,
        trainer_kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        """
        Fit the model to the data.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.
            neighbourhood (H3Neighbourhood): The neighbourhood to use.
                Should be intialized with the same regions.
            learning_rate (float, optional): Learning rate. Defaults to 0.001.
            trainer_kwargs (Optional[Dict[str, Any]], optional): Trainer kwargs.
                This is where the number of epochs can be set. Defaults to None.
        """
        import pytorch_lightning as pl

        trainer_kwargs = self._prepare_trainer_kwargs(trainer_kwargs)
        counts_df, dataloader, dataset = self._prepare_dataset(  # type: ignore
            regions_gdf, features_gdf, joint_gdf, neighbourhood, self._batch_size, shuffle=True
        )

        self._prepare_model(counts_df, learning_rate)

        trainer = pl.Trainer(**trainer_kwargs)
        trainer.fit(self._model, dataloader)
        self._is_fitted = True
        self._dataset = dataset

    def _prepare_model(self, counts_df: pd.DataFrame, learning_rate: float) -> None:
        if self._model is None:
            self._model = GeoVexModel(
                k_dim=len(counts_df.columns),
                radius=self._r,
                conv_layers=self._convolutional_layers,
                emb_size=self._embedding_size,
                learning_rate=learning_rate,
                conv_layer_size=self._convolutional_layer_size,
            )

    def _prepare_dataset(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
        neighbourhood: H3Neighbourhood,
        batch_size: Optional[int],
        shuffle: bool = True,
    ) -> tuple[pd.DataFrame, DataLoader, HexagonalDataset[T]]:
        counts_df = self._get_raw_counts(regions_gdf, features_gdf, joint_gdf)
        dataset: HexagonalDataset[T] = HexagonalDataset(
            counts_df,
            neighbourhood,
            neighbor_k_ring=self._r,
        )
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)
        return counts_df, dataloader, dataset

    def fit_transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
        neighbourhood: H3Neighbourhood,
        learning_rate: float = 0.001,
        trainer_kwargs: Optional[dict[str, Any]] = None,
    ) -> pd.DataFrame:
        """
        Fit the model to the data and create region embeddings.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.
            neighbourhood (H3Neighbourhood): The neighbourhood to use.
                Should be intialized with the same regions.
            negative_sample_k_distance (int, optional): Distance of negative samples. Defaults to 2.
            learning_rate (float, optional): Learning rate. Defaults to 0.001.
            trainer_kwargs (Optional[Dict[str, Any]], optional): Trainer kwargs. This is where the
                number of epochs can be set. Defaults to None.
        """
        self.fit(
            regions_gdf=regions_gdf,
            features_gdf=features_gdf,
            joint_gdf=joint_gdf,
            neighbourhood=neighbourhood,
            learning_rate=learning_rate,
            trainer_kwargs=trainer_kwargs,
        )
        assert self._dataset is not None  # for mypy
        return self._transform(dataset=self._dataset)

    def _get_raw_counts(
        self, regions_gdf: pd.DataFrame, features_gdf: pd.DataFrame, joint_gdf: pd.DataFrame
    ) -> pd.DataFrame:
        return super().transform(regions_gdf, features_gdf, joint_gdf).astype(np.float32)

    def _check_is_fitted(self) -> None:
        if not self._is_fitted or self._model is None:
            raise ModelNotFitException("Model not fitted. Call fit() or fit_transform() first.")

    def _prepare_trainer_kwargs(self, trainer_kwargs: Optional[dict[str, Any]]) -> dict[str, Any]:
        if trainer_kwargs is None:
            trainer_kwargs = {}
        if "max_epochs" not in trainer_kwargs:
            trainer_kwargs["max_epochs"] = 3
        return trainer_kwargs

    def _assert_feature_length(self, target_features: list[str], conv_layer_size: int) -> None:
        if len(target_features) < conv_layer_size:
            raise ValueError(
                f"The convolutional layers in GeoVex expect >= {conv_layer_size} features."
            )

    def save(self, path: Union[str, Any]) -> None:
        """
        Save the model to a directory.

        Args:
            path (Union[str, Any]): Path to the directory.
        """
        # embedder_config must match the constructor signature:
        # target_features: Union[list[str], OsmTagsFilter, GroupedOsmTagsFilter],
        # batch_size: Optional[int] = 32,
        # neighbourhood_radius: int = 4,
        # convolutional_layers: int = 2,
        # embedding_size: int = 32,
        # convolutional_layer_size: int = 256,
        embedder_config = {
            "target_features": self.expected_output_features.to_json(),
            "batch_size": self._batch_size,
            "neighbourhood_radius": self._r,
            "convolutional_layers": self._convolutional_layers,
            "embedding_size": self._embedding_size,
            "convolutional_layer_size": self._convolutional_layer_size,
        }
        self._save(path, embedder_config)

    def _save(self, path: Union[str, Any], embedder_config: dict[str, Any]) -> None:
        if isinstance(path, str):
            path = Path(path)

        self._check_is_fitted()

        path.mkdir(parents=True, exist_ok=True)

        # save model and config
        self._model.save(path / "model.pt")  # type: ignore
        # combine model config and embedder config
        model_config = self._model.get_config()  # type: ignore

        config = {
            "model_config": model_config,
            "embedder_config": embedder_config,
        }

        with (path / "config.json").open("w") as f:
            json.dump(config, f, ensure_ascii=False, indent=4)

    @classmethod
    def load(cls, path: Union[Path, str]) -> "GeoVexEmbedder":
        """
        Load the model from a directory.

        Args:
            path (Union[Path, str]): Path to the directory.
            model_module (type[ModelT]): Model class.

        Returns:
            GeoVexEmbedder: GeoVexEmbedder object.
        """
        return cls._load(path, GeoVexModel)

    @classmethod
    def _load(cls, path: Union[Path, str], model_module: type[ModelT]) -> "GeoVexEmbedder":
        if isinstance(path, str):
            path = Path(path)
        with (path / "config.json").open("r") as f:
            config = json.load(f)

        config["embedder_config"]["target_features"] = json.loads(
            config["embedder_config"]["target_features"]
        )
        embedder = cls(**config["embedder_config"])
        model_path = path / "model.pt"
        model = model_module.load(model_path, **config["model_config"])
        embedder._model = model
        embedder._is_fitted = True
        return embedder



"""
OSM Way loader.

This module contains osm loader implementation for ways based on OSMnx.
"""

import logging
from enum import Enum
from typing import Any, Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely.geometry as shpg
from functional import seq
from tqdm.auto import tqdm

from srai._optional import import_optional_dependencies
from srai.constants import FEATURES_INDEX, GEOMETRY_COLUMN, WGS84_CRS
from srai.exceptions import LoadedDataIsEmptyException
from srai.loaders import Loader

from . import constants

logger = logging.getLogger(__name__)


class OSMNetworkType(str, Enum):
    """
    Type of the street network.

    See [1] for more details.

    References:
        1. https://osmnx.readthedocs.io/en/stable/osmnx.html#osmnx.graph.graph_from_place
    """

    ALL_PRIVATE = "all_private"
    ALL = "all"
    BIKE = "bike"
    DRIVE = "drive"
    DRIVE_SERVICE = "drive_service"
    WALK = "walk"


class OSMWayLoader(Loader):
    """
    OSMWayLoader downloads road infrastructure from OSM.

    OSMWayLoader loader is a wrapper for the `osmnx.graph_from_polygon()`
    and `osmnx.graph_to_gdfs()` that simplifies obtaining the road infrastructure data
    from OpenStreetMap. As the OSM data is often noisy, it can also take an opinionated approach
    to preprocessing it, with standardisation in mind - e.g. unification of units,
    discarding non-wiki values and rounding them.
    """

    def __init__(
        self,
        network_type: Union[OSMNetworkType, str],
        contain_within_area: bool = False,
        preprocess: bool = True,
        wide: bool = True,
        metadata: bool = False,
        osm_way_tags: dict[str, list[str]] = constants.OSM_WAY_TAGS,
    ) -> None:
        """
        Init OSMWayLoader.

        Args:
            network_type (Union[NetworkType, str]):
                Type of the network to download.
            contain_within_area (bool): defaults to False
                Whether to remove the roads that have one of their nodes outside of the given area.
            preprocess (bool): defaults to True
                Whether to preprocess the data.
            wide (bool): defaults to True
                Whether to return the roads in wide format.
            metadata (bool): defaults to False
                Whether to return metadata for roads.
            osm_way_tags (List[str]): defaults to constants.OSM_WAY_TAGS
                Dict of tags to take into consideration during computing.
        """
        import_optional_dependencies(dependency_group="osm", modules=["osmnx"])

        self.network_type = network_type
        self.contain_within_area = contain_within_area
        self.preprocess = preprocess
        self.wide = wide
        self.metadata = metadata
        self.osm_keys = list(osm_way_tags.keys())
        self.osm_tags_flat = (
            seq(osm_way_tags.items())
            .flat_map(lambda x: [f"{x[0]}-{v}" if x[0] not in ("oneway") else x[0] for v in x[1]])
            .distinct()
            .to_list()
        )

    def load(self, area: gpd.GeoDataFrame) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
        """
        Load road infrastructure for a given GeoDataFrame.

        Args:
            area (gpd.GeoDataFrame): (Multi)Polygons for which to download road infrastructure data.

        Raises:
            ValueError: If provided GeoDataFrame has no crs defined.
            ValueError: If provided GeoDataFrame is empty.
            TypeError: If provided geometries are not of type Polygon or MultiPolygon.
            LoadedDataIsEmptyException: If none of the supplied area polygons contains
                any road infrastructure data.

        Returns:
            Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]: Road infrastructure as (intersections, roads)
        """
        import osmnx as ox

        ox.settings.useful_tags_way = constants.OSMNX_WAY_KEYS
        ox.settings.timeout = constants.OSMNX_TIMEOUT

        if area.empty:
            raise ValueError("Provided `area` GeoDataFrame is empty.")

        gdf_wgs84 = area.to_crs(crs=WGS84_CRS)

        gdf_nodes_raw, gdf_edges_raw = self._graph_from_gdf(gdf_wgs84)
        if gdf_edges_raw.empty or gdf_edges_raw.empty:
            raise LoadedDataIsEmptyException(
                "It can happen when there is no road infrastructure in the given area."
            )

        gdf_edges = self._explode_cols(gdf_edges_raw)

        if self.preprocess:
            gdf_edges = self._preprocess(gdf_edges)

        if self.wide:
            gdf_edges = self._to_wide(gdf_edges_raw, gdf_edges)

        gdf_edges = self._unify_index_and_columns_names(gdf_edges)

        return gdf_nodes_raw, gdf_edges

    def _graph_from_gdf(self, gdf: gpd.GeoDataFrame) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
        """
        Obtain the raw road infrastructure data from OSM.

        Args:
            gdf (gpd.GeoDataFrame): (Multi)Polygons for which to download road infrastructure data.

        Returns:
            Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]: Road infrastructure as (intersections, roads)
        """
        nodes = []
        edges = []
        for polygon in tqdm(gdf["geometry"], desc="Downloading graphs", leave=False):
            gdf_n, gdf_e = self._try_graph_from_polygon(polygon)

            if not gdf_e.empty and not self.contain_within_area:
                # perform cleaning of edges outside of an area that were incorrectly added,
                # it occures when two nodes outside of an area happen to be connected by an edge
                gdf_e = gdf_e.sjoin(
                    gpd.GeoDataFrame(geometry=[polygon], crs=WGS84_CRS),
                    how="inner",
                    predicate="intersects",
                ).drop(columns="index_right")

            nodes.append(gdf_n)
            edges.append(gdf_e)

        gdf_nodes = pd.concat(nodes, axis=0)
        gdf_edges = pd.concat(edges, axis=0)

        # remove duplicates, cannot use drop_duplicates()
        # because some columns contain unhashable type `list`
        gdf_nodes = gdf_nodes[~gdf_nodes.astype(str).duplicated()]
        gdf_edges = gdf_edges[~gdf_edges.astype(str).duplicated()]

        return gdf_nodes, gdf_edges

    def _try_graph_from_polygon(
        self, polygon: Union[shpg.Polygon, shpg.MultiPolygon]
    ) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
        """
        Try obtaining the raw road infrastructure data from OSM for a single polygon using `osmnx`.

        If `osmnx` fails, then just return the empty result.

        Args:
            polygon (Union[shapely.geometry.Polygon, shapely.geometry.MultiPolygon]):
                Polygon for which to download road infrastructure data.

        Returns:
            Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]: Road infrastructure as (intersections, roads)
        """
        import osmnx as ox
        from packaging import version

        current_osmnx_version = version.parse(ox.__version__)

        if current_osmnx_version >= version.parse("1.6.0"):
            from osmnx._errors import InsufficientResponseError

            response_error = InsufficientResponseError
        else:
            from osmnx._errors import EmptyOverpassResponse

            response_error = EmptyOverpassResponse

        try:
            return self._graph_from_polygon(polygon)
        except (response_error, ValueError):
            return gpd.GeoDataFrame(), gpd.GeoDataFrame()

    def _graph_from_polygon(
        self, polygon: Union[shpg.Polygon, shpg.MultiPolygon]
    ) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
        """
        Obtain the raw road infrastructure data from OSM for a single polygon.

        Args:
            polygon (Union[shapely.geometry.Polygon, shapely.geometry.MultiPolygon]):
                Polygon for which to download road infrastructure data.

        Notes:
            * The road infrastructure graph is treated as an undirected graph.

        Returns:
            Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]: Road infrastructure as (intersections, roads)
        """
        import osmnx as ox
        from packaging import version

        current_osmnx_version = version.parse(ox.__version__)

        if current_osmnx_version >= version.parse("2.0.0"):
            G_directed = ox.graph.graph_from_polygon(
                polygon,
                network_type=self.network_type,
                retain_all=True,
                truncate_by_edge=(not self.contain_within_area),
            )
            G_undirected = ox.convert.to_undirected(G_directed)
        else:
            G_directed = ox.graph_from_polygon(
                polygon,
                network_type=self.network_type,
                retain_all=True,
                clean_periphery=True,
                truncate_by_edge=(not self.contain_within_area),
            )
            G_undirected = ox.utils_graph.get_undirected(G_directed)

        gdf_n, gdf_e = ox.graph_to_gdfs(G_undirected)

        return gdf_n, gdf_e

    def _explode_cols(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Explode lists in feature columns.

        Args:
            gdf (gpd.GeoDataFrame): Edges with columns to explode.

        Returns:
            gpd.GeoDataFrame: Edges with all of their columns exploded.
        """
        for col in self.osm_keys:
            if col not in gdf.columns:
                gdf = gdf.assign(**{col: None})
            gdf = gdf.explode(col)

        gdf["i"] = range(len(gdf))
        gdf.set_index("i", append=True, inplace=True)

        return gdf

    def _preprocess(
        self, gdf: gpd.GeoDataFrame, inplace: bool = False
    ) -> Optional[gpd.GeoDataFrame]:
        """
        Preprocess edges.

        Args:
            gdf (gpd.GeoDataFrame): Edges to preprocess.
            inplace (bool): defaults to False.

        Returns:
            gpd.GeoDataFrame: Edges with preprocessed features.
        """
        if not inplace:
            gdf = gdf.copy()

        max_osm_keys_str_len = max(map(len, self.osm_keys))
        for col in (pbar := tqdm(self.osm_keys, leave=False)):
            pbar.set_description(f"Preprocessing {col:{max_osm_keys_str_len}}")
            gdf[col] = gdf[col].apply(
                lambda x, c=col: self._sanitize_and_normalize(x, c)  # noqa: FURB111
            )

        return gdf if not inplace else None

    def _sanitize_and_normalize(self, x: Any, column_name: str) -> str:
        return self._normalize(self._sanitize(str(x), column_name), column_name)

    def _normalize(self, x: Any, column_name: str) -> str:
        try:
            if x is None:
                return "None"
            elif column_name == "lanes":
                x = min(x, 15)
            elif column_name == "maxspeed":
                if x <= 0:
                    x = 0
                elif x <= 5:
                    x = 5
                elif x <= 7:
                    x = 7
                elif x <= 10:
                    x = 10
                elif x <= 15:
                    x = 15
                else:
                    x = min(int(round(x / 10) * 10), 200)
            elif column_name == "width":
                x = min(round(x * 2) / 2, 30.0)
        except Exception as e:
            logger.warning(
                f"{OSMWayLoader._normalize.__qualname__} | {column_name}: {x} - {type(x)} | {e}."
                " Returning 'None'"
            )
            return "None"

        return str(x)

    def _sanitize(self, x: Any, column_name: str) -> Any:
        if x in ("", "none", "None", np.nan, "nan", "NaN", None):
            return None

        try:
            if column_name == "lanes":
                x = int(float(x))
            elif column_name == "maxspeed":
                if x in ("signals", "variable"):
                    return None

                if x in constants.OSM_IMPLICIT_MAXSPEEDS:
                    x = constants.OSM_IMPLICIT_MAXSPEEDS[x]

                x = x.replace("km/h", "")
                if "mph" in x:
                    x = float(x.split("mph")[0].strip())
                    x = x * constants.MPH_TO_KMH
                x = float(x)
            elif column_name == "width":
                if x.endswith(("m", "meter")):
                    x = x.split("m")[0].strip()
                elif "'" in x:
                    x = float(x.split("'")[0].strip())
                    x = x * constants.INCHES_TO_METERS
                elif x.endswith("ft"):
                    x = float(x.split("ft")[0].strip())
                    x = x * constants.FEET_TO_METERS
                x = float(x)

        except Exception as e:
            logger.warning(
                f"{OSMWayLoader._sanitize.__qualname__} | {column_name}: {x} - {type(x)} | {e}."
                " Returning None"
            )
            return None

        return x

    def _to_wide(self, gdf: gpd.GeoDataFrame, gdf_exploded: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Convert edges in long format to wide.

        Args:
            gdf (gpd.GeoDataFrame): original edges.
            gdf_exploded (gpd.GeoDataFrame): edges with columns after feature explosion.

        Returns:
            gpd.GeoDataFrame: Edges in wide format.
        """
        gdf_edges_wide = (
            pd.get_dummies(gdf_exploded[self.osm_keys], prefix_sep="-")
            .droplevel(3)
            .groupby(level=[0, 1, 2])
            .max()
            .astype(np.uint8)
            .reindex(columns=self.osm_tags_flat, fill_value=0)
            .astype(np.uint8)
        )

        osm_keys_to_drop = [k for k in self.osm_keys if k in gdf.columns]
        gdf_edges_wide = gpd.GeoDataFrame(
            pd.concat(
                [
                    gdf.drop(columns=osm_keys_to_drop),
                    gdf_edges_wide,
                ],
                axis=1,
            ),
            crs=WGS84_CRS,
        )

        return gdf_edges_wide

    def _unify_index_and_columns_names(self, gdf_edges: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Make naming of index and columns consistent.

        Args:
            gdf_edges (gpd.GeoDataFrame): Edges to unify

        Returns:
            gpd.GeoDataFrame: Edges with unified index and columns names
        """
        gdf = gdf_edges.reset_index().drop(columns=["u", "v"])
        gdf.index.rename(FEATURES_INDEX, inplace=True)

        reindex_columns = constants.METADATA_COLUMNS if self.metadata else []
        reindex_columns += self.osm_tags_flat if self.wide else self.osm_keys
        reindex_columns += [GEOMETRY_COLUMN]
        gdf = gdf.reindex(columns=reindex_columns)

        return gdf



"""
OSM tile loader.

This module implements downloading tiles from given OSM tile server.
"""

from collections.abc import Iterable
from io import BytesIO
from pathlib import Path
from typing import Any, Optional, Union
from urllib.parse import urljoin

import geopandas as gpd
import pandas as pd
import requests
from shapely.geometry.base import BaseGeometry

from srai._optional import import_optional_dependencies
from srai.loaders.osm_loaders._base import prepare_area_gdf_for_loader
from srai.regionalizers.slippy_map_regionalizer import SlippyMapRegionalizer

from .osm_tile_data_collector import (
    DataCollector,
    DataCollectorType,
    InMemoryDataCollector,
    get_collector,
)


class OSMTileLoader:
    """
    OSM Tile Loader.

    Download raster tiles from user specified tile server, like listed in [1]. Loader finds x, y
    coordinates [2] for specified area and downloads tiles. Address is built with schema
    {tile_server_url}/{zoom}/{x}/{y}.{resource_type}

    References:
        1. https://wiki.openstreetmap.org/wiki/Raster_tile_providers
        2. https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
    """

    def __init__(
        self,
        tile_server_url: str,
        zoom: int,
        verbose: bool = False,
        resource_type: str = "png",
        auth_token: Optional[str] = None,
        data_collector: Optional[Union[str, DataCollector]] = None,
        storage_path: Optional[Union[str, Path]] = None,
    ) -> None:
        """
        Initialize TileLoader.

        Args:
            tile_server_url (str): url of tile server, without z, x, y parameters
            zoom (int): zoom level [1]
            verbose (bool, optional): should print logs. Defaults to False.
            resource_type (str, optional): file extension. Added to the end of url.
                Defaults to "png".
            auth_token (str, optional): auth token. Added as access_token parameter
                to request. Defaults to None.
            data_collector (Union[str, DataCollector], optional): DataCollector object or
            enum defining default collector. If None uses InMemoryDataCollector. Defaults to None.
            If `return` uses  InMemoryDataCollector
            If `save` uses  SavingDataCollector
            storage_path (Union[str, Path], optional): path to save data,
                used with SavingDataCollector. Defaults to None.

        References:
            1. https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
        """
        import_optional_dependencies(dependency_group="osm", modules=["PIL"])
        self.zoom = zoom
        self.verbose = verbose
        self.resource_type = resource_type
        self.base_url = urljoin(tile_server_url, "{0}/{1}/{2}." + resource_type)
        self.auth_token = auth_token
        self.save_path = storage_path
        self.data_collector = (
            self._get_collector(data_collector)
            if data_collector is not None
            else InMemoryDataCollector()
        )
        self.regionalizer = SlippyMapRegionalizer(zoom=self.zoom)

    def _get_collector(
        self, storage_strategy: Union[str, DataCollectorType, DataCollector]
    ) -> DataCollector:
        if isinstance(storage_strategy, str):
            return get_collector(
                storage_strategy, save_path=self.save_path, file_extension=self.resource_type
            )
        return storage_strategy

    def load(
        self,
        area: Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame],
    ) -> gpd.GeoDataFrame:
        """
        Return all tiles of region.

        Args:
            area (Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame]):
                Area for which to download objects.

        Returns:
            gpd.GeoDataFrame: Pandas of tiles for each region in area transformed by DataCollector
        """
        area_wgs84 = prepare_area_gdf_for_loader(area)
        regions = self.regionalizer.transform(gdf=area_wgs84)
        regions["tile"] = regions.apply(self._get_tile_for_area, axis=1)
        return regions

    def _get_tile_for_area(self, row: pd.Series) -> Any:
        idx = row.name
        return self.get_tile_by_x_y(row["x"], row["y"], idx=idx)

    def get_tile_by_x_y(self, x: int, y: int, idx: Any = None) -> Any:
        """
        Download single tile from tile server. Return tile processed by DataCollector.

        Args:
            x(int): x tile coordinate
            y(int): y tile coordinate
            idx (Any): id of tile, if non created as x_y_self.zoom
        """
        from PIL import Image

        if idx is None:
            idx = f"{x}_{y}_{self.zoom}"
        url = self.base_url.format(self.zoom, x, y)
        if self.verbose:
            print(f"Getting tile from url: {url}")
        content = requests.get(url, params=dict(access_token=self.auth_token)).content
        tile = Image.open(BytesIO(content))
        return self.data_collector.store(idx, tile)



"""H3 shapely conversion tests."""

from pathlib import Path
from typing import Any, Callable
from unittest import TestCase

import geopandas as gpd
import pytest

from srai.h3 import shapely_geometry_to_h3
from tests.h3.conftest import _gdf_noop, _gdf_to_geometry_list, _gdf_to_geoseries
from tests.regionalizers.test_h3_regionalizer import H3_RESOLUTION

ut = TestCase()


@pytest.mark.parametrize(
    "geometry_fixture, resolution, expected_h3_cells_fixture",
    [
        ("gdf_single_point", 10, "expected_point_h3_index"),
        ("gdf_multipolygon", H3_RESOLUTION, "expected_unbuffered_h3_indexes"),
        ("gdf_polygons", H3_RESOLUTION, "expected_unbuffered_h3_indexes"),
    ],
)  # type: ignore
@pytest.mark.parametrize(
    "geometry_parser_function",
    [_gdf_noop, _gdf_to_geoseries, _gdf_to_geometry_list],
)  # type: ignore
def test_shapely_geometry_to_h3_unbuffered(
    geometry_fixture: str,
    resolution: int,
    expected_h3_cells_fixture: str,
    geometry_parser_function: Callable[[gpd.GeoDataFrame], Any],
    request: pytest.FixtureRequest,
) -> None:
    """Test checks if conversion from shapely to h3 works."""
    geometry = request.getfixturevalue(geometry_fixture)
    expected_h3_cells = request.getfixturevalue(expected_h3_cells_fixture)

    parsed_geometry = geometry_parser_function(geometry)
    h3_cells = shapely_geometry_to_h3(
        geometry=parsed_geometry, h3_resolution=resolution, buffer=False
    )
    ut.assertCountEqual(h3_cells, expected_h3_cells)


@pytest.mark.parametrize(
    "geometry_fixture, resolution, expected_h3_cells_fixture",
    [
        ("gdf_single_point", 10, "expected_point_h3_index"),
        ("gdf_multipolygon", H3_RESOLUTION, "expected_h3_indexes"),
        ("gdf_polygons", H3_RESOLUTION, "expected_h3_indexes"),
    ],
)  # type: ignore
@pytest.mark.parametrize(
    "geometry_parser_function",
    [_gdf_noop, _gdf_to_geoseries, _gdf_to_geometry_list],
)  # type: ignore
def test_shapely_geometry_to_h3_buffered(
    geometry_fixture: str,
    resolution: int,
    expected_h3_cells_fixture: str,
    geometry_parser_function: Callable[[gpd.GeoDataFrame], Any],
    request: pytest.FixtureRequest,
) -> None:
    """Test checks if conversion from shapely to h3 with buffer works."""
    geometry = request.getfixturevalue(geometry_fixture)
    expected_h3_cells = request.getfixturevalue(expected_h3_cells_fixture)

    parsed_geometry = geometry_parser_function(geometry)
    h3_cells = shapely_geometry_to_h3(
        geometry=parsed_geometry, h3_resolution=resolution, buffer=True
    )
    ut.assertCountEqual(h3_cells, expected_h3_cells)


def test_full_coverage() -> None:
    """Test if h3 coverage works if geometry is smaller than single cell."""
    gdf = gpd.read_file(Path(__file__).parent / "test_files" / "buildings.geojson")

    intersections = {}
    for geom, osm_id in zip(gdf.geometry, gdf.id):
        intersections[osm_id] = shapely_geometry_to_h3(geom, h3_resolution=10)

    assert all(len(value) > 0 for value in intersections.values())
    assert len(intersections["way/843232154"]) == 1
    assert "8a2ab5760167fff" in intersections["way/843232154"]



"""Tests for SlippyMapRegionalizer class."""

from contextlib import nullcontext as does_not_raise
from typing import Any

import geopandas as gpd
import pytest
from shapely.geometry import Polygon

from srai.constants import WGS84_CRS
from srai.regionalizers import SlippyMapRegionalizer

ZOOM = 11


@pytest.fixture  # type: ignore
def gdf() -> gpd.GeoDataFrame:
    """GeoDataFrame approximating Warsaw bounds."""
    polygon = Polygon(
        [
            (16.8073393, 51.1389477),
            (17.0278673, 51.0426754),
            (17.1762192, 51.1063195),
            (16.9580276, 51.2093551),
            (16.8073393, 51.1389477),
        ],
    )
    gdf = gpd.GeoDataFrame({"geometry": [polygon]}, crs=WGS84_CRS)
    return gdf


@pytest.fixture  # type: ignore
def regionalizer() -> SlippyMapRegionalizer:
    """Regionalizer fixture."""
    return SlippyMapRegionalizer(zoom=ZOOM)


def test_transform(regionalizer: SlippyMapRegionalizer, gdf: gpd.GeoDataFrame) -> None:
    """Test returned regions."""
    regions = regionalizer.transform(gdf)

    assert regions.shape[0] == 6, f"Invalid length {regions.shape[0]}"
    for x, y in zip([1120, 1119, 1120, 1121, 1120, 1121], [683, 684, 684, 684, 685, 685]):
        expected_id = f"{x}_{y}_{ZOOM}"
        assert expected_id in regions.index, f"{expected_id} not in index but expected"
        assert regions.loc[expected_id]["x"] == x
        assert regions.loc[expected_id]["y"] == y
        assert regions.loc[expected_id]["z"] == ZOOM


@pytest.mark.parametrize(  # type: ignore
    "z, expectation",
    [
        (-1, pytest.raises(ValueError)),
        (0, does_not_raise()),
        (19, does_not_raise()),
        (20, pytest.raises(ValueError)),
    ],
)
def test_zoom_check(z: int, expectation: Any) -> None:
    """Test value checks."""
    with expectation:
        SlippyMapRegionalizer(zoom=z)


def test_coordinates_cast(regionalizer: SlippyMapRegionalizer) -> None:
    """Test if coordinates_to_x_y gives proper x and y value."""
    # given
    latitude, longitude = 51, 16.8
    regionalizer.zoom = 10

    # when
    x, y = regionalizer._coordinates_to_x_y(latitude=latitude, longitude=longitude)

    # then
    assert x == 559
    assert y == 342


def test_x_y_to_coordinates_should_be_inverse_to_coordinates_to_x_y(
    regionalizer: SlippyMapRegionalizer,
) -> None:
    """Test if `x_y_to_coordinates` is reversible with `coordinates_to_x_y`."""
    # given
    x, y = 50, 100

    # when
    latitude, longitude = regionalizer._x_y_to_coordinates(x, y)
    x_reverse, y_reverse = regionalizer._coordinates_to_x_y(latitude, longitude)

    # then
    assert x_reverse == x
    assert y_reverse == y



"""Constants for geovex tests."""

from srai.loaders.osm_loaders.filters import HEX2VEC_FILTER

EMBEDDING_SIZE = 32
TRAINER_KWARGS = {"max_epochs": 1, "accelerator": "cpu", "deterministic": True}

PREDEFINED_TEST_CASES = [
    {
        "test_case_name": "wro_9",
        "root_region_index": "891e2040887ffff",
        "region_gen_radius": 2,
        "model_radius": 2,
        "seed": 42,
        "tags": {
            t: HEX2VEC_FILTER[t]
            for t in [
                "building",
                "amenity",
            ]
        },
        "convolutional_layer_size": 32,
        "num_layers": 2,
    },
    {
        "test_case_name": "AL_10",
        "root_region_index": "8a44ed664c27fff",
        "region_gen_radius": 3,
        "model_radius": 3,
        "seed": 5555,
        "tags": {
            t: HEX2VEC_FILTER[t]
            for t in [
                "building",
                "amenity",
            ]
        },
        "convolutional_layer_size": 32,
        "num_layers": 2,
    },
]



"""Tests for neigbourhoods with no regions."""

import pytest

from srai.neighbourhoods import H3Neighbourhood


@pytest.mark.parametrize(  # type: ignore
    "index,expected,expected_with_include_center",
    [
        (
            "811e3ffffffffff",
            {
                "811f3ffffffffff",
                "811fbffffffffff",
                "811ebffffffffff",
                "811efffffffffff",
                "811e7ffffffffff",
                "811f7ffffffffff",
            },
            {
                "811f3ffffffffff",
                "811fbffffffffff",
                "811ebffffffffff",
                "811efffffffffff",
                "811e7ffffffffff",
                "811f7ffffffffff",
                "811e3ffffffffff",
            },
        ),
        (
            "831f0bfffffffff",
            {
                "831f08fffffffff",
                "831f0afffffffff",
                "831e24fffffffff",
                "831e25fffffffff",
                "831f56fffffffff",
                "831f09fffffffff",
            },
            {
                "831f08fffffffff",
                "831f0afffffffff",
                "831e24fffffffff",
                "831e25fffffffff",
                "831f56fffffffff",
                "831f09fffffffff",
                "831f0bfffffffff",
            },
        ),
        (
            "882baa7b69fffff",
            {
                "882baa4e97fffff",
                "882baa7b6dfffff",
                "882baa7b61fffff",
                "882baa7b6bfffff",
                "882baa7b45fffff",
                "882baa4e93fffff",
            },
            {
                "882baa4e97fffff",
                "882baa7b6dfffff",
                "882baa7b61fffff",
                "882baa7b6bfffff",
                "882baa7b45fffff",
                "882baa4e93fffff",
                "882baa7b69fffff",
            },
        ),
    ],
)
def test_get_neighbours(
    index: str, expected: set[str], expected_with_include_center: set[str]
) -> None:
    """Test get_neighbours of H3Neighbourhood."""
    neighbourhood = H3Neighbourhood()
    assert neighbourhood.get_neighbours(index) == expected
    assert neighbourhood.get_neighbours(index, include_center=True) == expected_with_include_center

    neighbourhood_with_include_center = H3Neighbourhood(include_center=True)
    assert neighbourhood_with_include_center.get_neighbours(index) == expected_with_include_center
    assert neighbourhood_with_include_center.get_neighbours(index, include_center=False) == expected


@pytest.mark.parametrize(  # type: ignore
    "index,distance,expected,expected_with_include_center",
    [
        ("811e3ffffffffff", -2, set(), set()),
        ("811e3ffffffffff", -1, set(), set()),
        ("811e3ffffffffff", 0, set(), {"811e3ffffffffff"}),
        (
            "861f09b27ffffff",
            1,
            {
                "861f09b07ffffff",
                "861f09b2fffffff",
                "861f09b37ffffff",
                "861f7248fffffff",
                "861f7249fffffff",
                "861f724d7ffffff",
            },
            {
                "861f09b07ffffff",
                "861f09b2fffffff",
                "861f09b37ffffff",
                "861f7248fffffff",
                "861f7249fffffff",
                "861f724d7ffffff",
                "861f09b27ffffff",
            },
        ),
        (
            "861f09b27ffffff",
            2,
            {
                "861f0984fffffff",
                "861f0986fffffff",
                "861f09b07ffffff",
                "861f09b0fffffff",
                "861f09b17ffffff",
                "861f09b1fffffff",
                "861f09b2fffffff",
                "861f09b37ffffff",
                "861f09b77ffffff",
                "861f72487ffffff",
                "861f7248fffffff",
                "861f72497ffffff",
                "861f7249fffffff",
                "861f724afffffff",
                "861f724c7ffffff",
                "861f724d7ffffff",
                "861f724dfffffff",
                "861f724f7ffffff",
            },
            {
                "861f0984fffffff",
                "861f0986fffffff",
                "861f09b07ffffff",
                "861f09b0fffffff",
                "861f09b17ffffff",
                "861f09b1fffffff",
                "861f09b2fffffff",
                "861f09b37ffffff",
                "861f09b77ffffff",
                "861f72487ffffff",
                "861f7248fffffff",
                "861f72497ffffff",
                "861f7249fffffff",
                "861f724afffffff",
                "861f724c7ffffff",
                "861f724d7ffffff",
                "861f724dfffffff",
                "861f724f7ffffff",
                "861f09b27ffffff",
            },
        ),
    ],
)
def test_get_neighbours_up_to_distance(
    index: str, distance: int, expected: set[str], expected_with_include_center: set[str]
) -> None:
    """Test get_neighbours_up_to_distance of H3Neighbourhood."""
    neighbourhood = H3Neighbourhood()
    assert neighbourhood.get_neighbours_up_to_distance(index, distance) == expected
    assert (
        neighbourhood.get_neighbours_up_to_distance(index, distance, include_center=True)
        == expected_with_include_center
    )

    neighbourhood_with_include_center = H3Neighbourhood(include_center=True)
    assert (
        neighbourhood_with_include_center.get_neighbours_up_to_distance(index, distance)
        == expected_with_include_center
    )
    assert (
        neighbourhood_with_include_center.get_neighbours_up_to_distance(
            index, distance, include_center=False
        )
        == expected
    )


@pytest.mark.parametrize(  # type: ignore
    "index,distance,expected,expected_with_include_center",
    [
        ("811e3ffffffffff", -2, set(), set()),
        ("811e3ffffffffff", -1, set(), set()),
        ("811e3ffffffffff", 0, set(), {"811e3ffffffffff"}),
        (
            "862bac507ffffff",
            1,
            {
                "862bac50fffffff",
                "862bac517ffffff",
                "862bac51fffffff",
                "862bac527ffffff",
                "862bac52fffffff",
                "862bac537ffffff",
            },
            {
                "862bac50fffffff",
                "862bac517ffffff",
                "862bac51fffffff",
                "862bac527ffffff",
                "862bac52fffffff",
                "862bac537ffffff",
            },
        ),
        (
            "862bac507ffffff",
            2,
            {
                "862ba124fffffff",
                "862ba126fffffff",
                "862bac427ffffff",
                "862bac437ffffff",
                "862bac557ffffff",
                "862bac577ffffff",
                "862bac5a7ffffff",
                "862bac5afffffff",
                "862bacc8fffffff",
                "862bacc9fffffff",
                "862baccd7ffffff",
                "862baccdfffffff",
            },
            {
                "862ba124fffffff",
                "862ba126fffffff",
                "862bac427ffffff",
                "862bac437ffffff",
                "862bac557ffffff",
                "862bac577ffffff",
                "862bac5a7ffffff",
                "862bac5afffffff",
                "862bacc8fffffff",
                "862bacc9fffffff",
                "862baccd7ffffff",
                "862baccdfffffff",
            },
        ),
    ],
)
def test_get_neighbours_at_distance(
    index: str, distance: int, expected: set[str], expected_with_include_center: set[str]
) -> None:
    """Test get_neighbours_at_distance of H3Neighbourhood."""
    neighbourhood = H3Neighbourhood()
    assert neighbourhood.get_neighbours_at_distance(index, distance) == expected
    assert (
        neighbourhood.get_neighbours_at_distance(index, distance, include_center=True)
        == expected_with_include_center
    )

    neighbourhood_with_include_center = H3Neighbourhood(include_center=True)
    assert (
        neighbourhood_with_include_center.get_neighbours_at_distance(index, distance)
        == expected_with_include_center
    )
    assert (
        neighbourhood_with_include_center.get_neighbours_at_distance(
            index, distance, include_center=False
        )
        == expected
    )



"""Tests for GeoVex model."""

import os
from contextlib import nullcontext as does_not_raise
from pathlib import Path
from typing import Any, cast

import geopandas as gpd
import pandas as pd
import pytest
import torch
from pytorch_lightning import seed_everything

from srai.embedders.geovex.embedder import GeoVexEmbedder
from srai.embedders.geovex.model import GeoVexModel
from srai.neighbourhoods.h3_neighbourhood import H3Neighbourhood
from tests.embedders.geovex.constants import EMBEDDING_SIZE, PREDEFINED_TEST_CASES


@pytest.mark.parametrize(  # type: ignore
    "radius,expectation",
    [
        (1, pytest.raises(ValueError)),
        (2, does_not_raise()),
        (16, does_not_raise()),
    ],
)
def test_model_raises_with_incorrect_layer_sizes(radius: int, expectation: Any) -> None:
    """Test if GeoVexModel raises with incorrect layer_sizes."""
    with expectation:
        GeoVexModel(
            k_dim=256,
            radius=radius,
        )


@pytest.mark.parametrize(  # type: ignore
    "radius",
    [4, 9],
)
def test_layers_initialized_correctly(radius) -> None:
    """Test if GeoVexModel layers are initialized correctly."""
    k_dim = 512
    conv_layers = 3
    emb_size = 32
    conv_layer_size = 256
    model = GeoVexModel(
        k_dim=k_dim,
        radius=radius,
        conv_layers=conv_layers,
        emb_size=emb_size,
        conv_layer_size=conv_layer_size,
    )
    conv_sizes = [k_dim, *[conv_layer_size * 2**i for i in range(conv_layers)]]

    block_counter = 1
    for i, layer in enumerate(model.encoder):
        if 4 < i < (4 + conv_layers):
            # this is a conv block
            assert layer[0].conv.in_channels == conv_sizes[block_counter]
            block_counter += 1

    # check that they linear embedding dimension is correct
    assert model.encoder[-1].out_features == emb_size


def test_model_tensors() -> None:
    """Test if model batches are correctly parsed after seeding."""
    test_files_path = Path(__file__).parent / "test_files"
    for test_case in PREDEFINED_TEST_CASES:
        name = test_case["test_case_name"]
        seed = test_case["seed"]
        radius: int = test_case["model_radius"]  # type: ignore
        print(name, seed)

        regions_gdf = gpd.read_parquet(test_files_path / f"{name}_regions.parquet")
        features_gdf = gpd.read_parquet(test_files_path / f"{name}_features.parquet")
        joint_gdf = pd.read_parquet(test_files_path / f"{name}_joint.parquet")

        seed_everything(seed, workers=True)
        os.environ["PYTHONHASHSEED"] = str(seed)
        torch.use_deterministic_algorithms(True)

        neighbourhood = H3Neighbourhood(regions_gdf)
        target_features = [f"{st}_{t}" for st in test_case["tags"] for t in test_case["tags"][st]]  # type: ignore
        embedder = GeoVexEmbedder(
            target_features=target_features,
            batch_size=10,
            neighbourhood_radius=radius,
            embedding_size=EMBEDDING_SIZE,
            convolutional_layers=test_case["num_layers"],  # type: ignore
            convolutional_layer_size=test_case["convolutional_layer_size"],  # type: ignore
        )

        counts_df, dataloader, _ = embedder._prepare_dataset(
            regions_gdf, features_gdf, joint_gdf, neighbourhood, embedder._batch_size, shuffle=True
        )

        embedder._prepare_model(counts_df, 0.001)

        for _, param in cast(GeoVexModel, embedder._model).named_parameters():
            param.data.fill_(0.01)

        for i, batch in enumerate(dataloader):
            expected_batch = torch.load(test_files_path / f"{name}_batch_{i}.pt")
            torch.testing.assert_close(batch, expected_batch, rtol=0, atol=0)

            expected_encoder_forward_tensor = torch.load(
                test_files_path / f"{name}_encoder_forward_{i}.pt"
            )
            encoder_forward_tensor = cast(GeoVexModel, embedder._model).encoder.forward(batch)
            torch.testing.assert_close(encoder_forward_tensor, expected_encoder_forward_tensor)

            expected_forward_tensor = torch.load(test_files_path / f"{name}_forward_{i}.pt")
            forward_tensor = cast(GeoVexModel, embedder._model).forward(batch)
            torch.testing.assert_close(forward_tensor, expected_forward_tensor)

            expected_loss_tensor = torch.load(test_files_path / f"{name}_loss_{i}.pt")
            loss_tensor = cast(GeoVexModel, embedder._model).training_step(batch, i)
            torch.testing.assert_close(loss_tensor, expected_loss_tensor)



"""Optional dependencies tests."""

import sys
from contextlib import nullcontext as does_not_raise
from typing import Any

import geopandas as gpd
import pytest
from shapely.geometry import box

from srai._optional import ImportErrorHandle, import_optional_dependency
from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS


@pytest.fixture  # type: ignore
def optional_packages() -> list[str]:
    """Get a list with optional packages."""
    return [
        "quackosm",
        "osmnx",
        "overpass",
        "beautifulsoup4",
        "pymap3d",
        "haversine",
        "scipy",
        "spherical_geometry",
        "gtfs_kit",
        "folium",
        "mapclassify",
        "plotly",
        "kaleido",
        "pytorch-lightning",
        "torch",
    ]


@pytest.fixture(autouse=True)  # type: ignore
def cleanup_imports():
    """Clean imports."""
    yield
    sys.modules.pop("srai", None)


class PackageDiscarder:
    """Mock class for discarding list of packages."""

    def __init__(self) -> None:
        """Init mock class."""
        self.pkgnames: list[str] = []

    def find_spec(self, fullname, path, target=None) -> None:  # type: ignore
        """Throws ImportError if matching module."""
        if fullname in self.pkgnames:
            raise ImportError()


@pytest.fixture  # type: ignore
def no_optional_dependencies(monkeypatch, optional_packages):
    """Mock environment without optional dependencies."""
    d = PackageDiscarder()

    for package in optional_packages:
        sys.modules.pop(package, None)
        d.pkgnames.append(package)
    sys.meta_path.insert(0, d)
    yield
    sys.meta_path.remove(d)


def _test_voronoi() -> None:
    import geopandas as gpd
    from shapely.geometry import Point

    from srai.regionalizers import VoronoiRegionalizer

    seeds_gdf = gpd.GeoDataFrame(
        {
            GEOMETRY_COLUMN: [
                Point(17.014997869227177, 51.09919872601259),
                Point(16.935542631959215, 51.09380600286582),
                Point(16.900425, 51.1162552343),
                Point(16.932700, 51.166251),
            ]
        },
        index=[1, 2, 3, 4],
        crs=WGS84_CRS,
    )
    VoronoiRegionalizer(seeds=seeds_gdf)


def _test_plotting() -> None:
    from srai.plotting import folium_wrapper, plotly_wrapper

    folium_wrapper.plot_regions(_get_regions_gdf())
    plotly_wrapper.plot_regions(_get_regions_gdf(), return_plot=True)


def _test_torch() -> None:
    from srai.embedders import (
        GeoVexEmbedder,
        GTFS2VecEmbedder,
        Hex2VecEmbedder,
        Highway2VecEmbedder,
    )

    Highway2VecEmbedder()
    GTFS2VecEmbedder()
    Hex2VecEmbedder()
    GeoVexEmbedder(["a"] * 256)


def _test_osm() -> None:
    from srai.loaders import OSMOnlineLoader, OSMPbfLoader, OSMTileLoader, OSMWayLoader
    from srai.regionalizers import AdministrativeBoundaryRegionalizer

    AdministrativeBoundaryRegionalizer(2)
    OSMPbfLoader()
    OSMOnlineLoader()
    OSMWayLoader("drive")
    OSMTileLoader("https://tile.openstreetmap.de", 9)


def _test_gtfs() -> None:
    from srai.loaders import GTFSLoader

    GTFSLoader()


def _get_regions_gdf() -> gpd.GeoDataFrame:
    return gpd.GeoDataFrame(
        data={
            GEOMETRY_COLUMN: [
                box(
                    minx=0,
                    miny=0,
                    maxx=1,
                    maxy=1,
                )
            ],
            REGIONS_INDEX: [1],
        },
        crs=WGS84_CRS,
    )


@pytest.mark.parametrize(  # type: ignore
    "test_fn",
    [
        (_test_voronoi),
        (_test_plotting),
        (_test_torch),
        (_test_osm),
        (_test_gtfs),
    ],
)
def test_optional_available(test_fn):
    """Test if defined functions are working with optional packages."""
    test_fn()


@pytest.mark.usefixtures("no_optional_dependencies")
@pytest.mark.parametrize(  # type: ignore
    "test_fn",
    [
        (_test_voronoi),
        (_test_plotting),
        (_test_torch),
        (_test_osm),
        (_test_gtfs),
    ],
)
def test_optional_missing(test_fn):
    """Test if defined functions are failing without optional packages."""
    with pytest.raises(ImportError):
        test_fn()


@pytest.mark.usefixtures("no_optional_dependencies")  # type: ignore
@pytest.mark.parametrize(  # type: ignore
    "import_error,expectation",
    [
        (ImportErrorHandle.RAISE, pytest.raises(ImportError)),
        (ImportErrorHandle.WARN, pytest.warns(ImportWarning)),
        (ImportErrorHandle.IGNORE, does_not_raise()),
    ],
)
def test_optional_missing_error_handle(import_error: ImportErrorHandle, expectation: Any) -> None:
    """Test checks if import error handles are working."""
    with expectation:
        import_optional_dependency(dependency_group="test", module="_srai_test", error=import_error)



"""
Hex2Vec Embedder.

This module contains embedder from Hex2Vec paper[1].

References:
    [1] https://dl.acm.org/doi/10.1145/3486635.3491076
"""

import json
from pathlib import Path
from typing import Any, Optional, TypeVar, Union

import geopandas as gpd
import numpy as np
import pandas as pd

from srai._optional import import_optional_dependencies
from srai.embedders import CountEmbedder, ModelT
from srai.embedders.hex2vec.model import Hex2VecModel
from srai.embedders.hex2vec.neighbour_dataset import NeighbourDataset
from srai.exceptions import ModelNotFitException
from srai.loaders.osm_loaders.filters import GroupedOsmTagsFilter, OsmTagsFilter
from srai.neighbourhoods import Neighbourhood

T = TypeVar("T")


class Hex2VecEmbedder(CountEmbedder):
    """Hex2Vec Embedder."""

    DEFAULT_ENCODER_SIZES = [150, 75, 50]

    def __init__(
        self,
        encoder_sizes: Optional[list[int]] = None,
        expected_output_features: Optional[
            Union[list[str], OsmTagsFilter, GroupedOsmTagsFilter]
        ] = None,
    ) -> None:
        """
        Initialize Hex2VecEmbedder.

        Args:
            encoder_sizes (List[int], optional): Sizes of the encoder layers.
                The input layer size shouldn't be included - it's inferred from the data.
                The last element is the embedding size. Defaults to [150, 75, 50].
            expected_output_features
                (Union[List[str], OsmTagsFilter, GroupedOsmTagsFilter], optional):
                List of expected output features. Defaults to None.
        """
        super().__init__(
            expected_output_features=expected_output_features, count_subcategories=True
        )
        import_optional_dependencies(
            dependency_group="torch", modules=["torch", "pytorch_lightning"]
        )
        if encoder_sizes is None:
            encoder_sizes = Hex2VecEmbedder.DEFAULT_ENCODER_SIZES
        self._assert_encoder_sizes_correct(encoder_sizes)
        self._encoder_sizes = encoder_sizes
        self._model: Optional[Hex2VecModel] = None
        self._is_fitted = False

    def transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:
        """
        Create region embeddings.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Returns:
            pd.DataFrame: Embedding and geometry index for each region in regions_gdf.

        Raises:
            ValueError: If features_gdf is empty and self.expected_output_features is not set.
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        import torch

        self._check_is_fitted()
        counts_df = self._get_raw_counts(regions_gdf, features_gdf, joint_gdf)
        counts_tensor = torch.from_numpy(counts_df.values)
        embeddings = self._model(counts_tensor).detach().numpy()  # type: ignore
        return pd.DataFrame(embeddings, index=counts_df.index)

    def fit(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
        neighbourhood: Neighbourhood[T],
        negative_sample_k_distance: int = 2,
        batch_size: int = 32,
        learning_rate: float = 0.001,
        trainer_kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        """
        Fit the model to the data.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.
            neighbourhood (Neighbourhood[T]): The neighbourhood to use.
                Should be intialized with the same regions.
            negative_sample_k_distance (int, optional): When sampling negative samples,
                sample from a distance > k. Defaults to 2.
            batch_size (int, optional): Batch size. Defaults to 32.
            learning_rate (float, optional): Learning rate. Defaults to 0.001.
            trainer_kwargs (Optional[Dict[str, Any]], optional): Trainer kwargs. Defaults to None.

        Raises:
            ValueError: If features_gdf is empty and self.expected_output_features is not set.
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
            ValueError: If negative_sample_k_distance < 2.
        """
        import pytorch_lightning as pl
        from torch.utils.data import DataLoader

        trainer_kwargs = self._prepare_trainer_kwargs(trainer_kwargs)

        counts_df = self._get_raw_counts(regions_gdf, features_gdf, joint_gdf)

        if self.expected_output_features is None:  # type: ignore[has-type]
            self.expected_output_features = pd.Series(counts_df.columns)

        num_features = len(self.expected_output_features)  # type: ignore[arg-type]
        self._model = Hex2VecModel(
            layer_sizes=[num_features, *self._encoder_sizes], learning_rate=learning_rate
        )
        dataset = NeighbourDataset(counts_df, neighbourhood, negative_sample_k_distance)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

        trainer = pl.Trainer(**trainer_kwargs)
        trainer.fit(self._model, dataloader)
        self._is_fitted = True

    def fit_transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
        neighbourhood: Neighbourhood[T],
        negative_sample_k_distance: int = 2,
        batch_size: int = 32,
        learning_rate: float = 0.001,
        trainer_kwargs: Optional[dict[str, Any]] = None,
    ) -> pd.DataFrame:
        """
        Fit the model to the data and return the embeddings.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.
            neighbourhood (Neighbourhood[T]): The neighbourhood to use.
                Should be intialized with the same regions.
            negative_sample_k_distance (int, optional): When sampling negative samples,
                sample from a distance > k. Defaults to 2.
            batch_size (int, optional): Batch size. Defaults to 32.
            learning_rate (float, optional): Learning rate. Defaults to 0.001.
            trainer_kwargs (Optional[Dict[str, Any]], optional): Trainer kwargs. Defaults to None.

        Returns:
            pd.DataFrame: Region embeddings.

        Raises:
            ValueError: If features_gdf is empty and self.expected_output_features is not set.
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
            ValueError: If negative_sample_k_distance < 2.
        """
        self.fit(
            regions_gdf,
            features_gdf,
            joint_gdf,
            neighbourhood,
            negative_sample_k_distance,
            batch_size,
            learning_rate,
            trainer_kwargs,
        )
        return self.transform(regions_gdf, features_gdf, joint_gdf)

    def _prepare_trainer_kwargs(self, trainer_kwargs: Optional[dict[str, Any]]) -> dict[str, Any]:
        if trainer_kwargs is None:
            trainer_kwargs = {}
        if "max_epochs" not in trainer_kwargs:
            trainer_kwargs["max_epochs"] = 3
        return trainer_kwargs

    def _get_raw_counts(
        self, regions_gdf: pd.DataFrame, features_gdf: pd.DataFrame, joint_gdf: pd.DataFrame
    ) -> pd.DataFrame:
        return super().transform(regions_gdf, features_gdf, joint_gdf).astype(np.float32)

    def _check_is_fitted(self) -> None:
        if not self._is_fitted or self._model is None:
            raise ModelNotFitException("Model not fitted. Call fit() or fit_transform() first.")

    def _assert_encoder_sizes_correct(self, encoder_sizes: list[int]) -> None:
        if len(encoder_sizes) < 1:
            raise ValueError("Encoder sizes must have at least one element - embedding size.")
        if any(size <= 0 for size in encoder_sizes):
            raise ValueError("Encoder sizes must be positive integers.")

    def _save(self, path: Union[Path, str], embedder_config: dict[str, Any]) -> None:
        if isinstance(path, str):
            path = Path(path)

        self._check_is_fitted()

        path.mkdir(parents=True, exist_ok=True)

        self._model.save(path / "model.pt")  # type: ignore

        config = {
            "model_config": self._model.get_config(),  # type: ignore
            "embedder_config": embedder_config,
        }
        with (path / "config.json").open("w") as f:
            json.dump(config, f, ensure_ascii=False, indent=4)

    def save(self, path: Union[Path, str]) -> None:
        """
        Save the model to a directory.

        Args:
            path (Path): Path to the directory.
        """
        embedder_config = {
            "encoder_sizes": self._encoder_sizes,
            "expected_output_features": (
                self.expected_output_features.tolist()
                if self.expected_output_features is not None
                else None
            ),
        }
        self._save(path, embedder_config)

    @classmethod
    def _load(cls, path: Union[Path, str], model_module: type[ModelT]) -> "Hex2VecEmbedder":
        if isinstance(path, str):
            path = Path(path)

        with (path / "config.json").open("r") as f:
            config = json.load(f)
        embedder = cls(**config["embedder_config"])
        model_path = path / "model.pt"
        model = model_module.load(model_path, **config["model_config"])
        embedder._model = model
        embedder._is_fitted = True
        return embedder

    @classmethod
    def load(cls, path: Union[Path, str]) -> "Hex2VecEmbedder":
        """
        Load the model from a directory.

        Args:
            path (Path): Path to the directory.

        Returns:
            Hex2VecEmbedder: The loaded embedder.
        """
        return cls._load(path, Hex2VecModel)



"""
Voronoi Regionalizer.

This module contains voronoi regionalizer implementation.
"""

from typing import TYPE_CHECKING, Optional, Union

import geopandas as gpd
from shapely.geometry import Point, box

from srai._optional import import_optional_dependencies
from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS
from srai.regionalizers import Regionalizer

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Hashable


class VoronoiRegionalizer(Regionalizer):
    """
    VoronoiRegionalizer.

    Voronoi [1] regionalizer allows the given geometries to be divided
    into Thiessen polygons using geometries that are the seeds. To
    minimize distortions tessellation will be performed on a sphere
    using SphericalVoronoi [2] function from scipy library.

    References:
        1. https://en.wikipedia.org/wiki/Voronoi_diagram
        2. https://docs.scipy.org/doc/scipy-1.9.2/reference/generated/scipy.spatial.SphericalVoronoi.html
    """  # noqa: W505, E501

    def __init__(
        self,
        seeds: Union[gpd.GeoDataFrame, list[Point]],
        max_meters_between_points: int = 10_000,
        num_of_multiprocessing_workers: int = -1,
        multiprocessing_activation_threshold: Optional[int] = None,
    ) -> None:
        """
        Init VoronoiRegionalizer.

        All (multi)polygons from seeds GeoDataFrame will be transformed to their centroids,
        because scipy function requires only points as an input.

        Args:
            seeds (Union[gpd.GeoDataFrame, List[Point]]): List of points or a GeoDataFrame
                with seeds for creating a tessellation. Every non-point geometry will be mapped
                to a centroid. Minimum 4 seeds are required. Seeds cannot lie on a single arc.
                Empty seeds will be removed.
            max_meters_between_points (int): Maximal distance in meters between two points
                in a resulting polygon. Higher number results lower resolution of a polygon.
            num_of_multiprocessing_workers (int): Number of workers used for
                multiprocessing. Defaults to `-1` which results in a total number of available
                cpu threads. `0` and `1` values disable multiprocessing.
                Similar to `n_jobs` parameter from `scikit-learn` library.
            multiprocessing_activation_threshold (int, optional): Number of seeds required to start
                processing on multiple processes. Activating multiprocessing for a small
                amount of points might not be feasible. Defaults to 100.

        Raises:
            ValueError: If any seed is duplicated.
            ValueError: If less than 4 seeds are provided.
            ValueError: If provided seeds geodataframe has no crs defined.
            ValueError: If any seed is outside WGS84 coordinates domain.
        """
        import_optional_dependencies(
            dependency_group="voronoi",
            modules=["haversine", "pymap3d", "scipy", "spherical_geometry"],
        )
        self.region_ids: list[Hashable] = []
        self.seeds: list[Point] = []

        if isinstance(seeds, gpd.GeoDataFrame):
            from ._spherical_voronoi import _parse_geodataframe_seeds

            self.seeds, self.region_ids = _parse_geodataframe_seeds(seeds)
        else:
            self.seeds = seeds
            self.region_ids = list(range(len(seeds)))

        self.max_meters_between_points = max_meters_between_points
        self.num_of_multiprocessing_workers = num_of_multiprocessing_workers
        self.multiprocessing_activation_threshold = multiprocessing_activation_threshold

        if len(self.seeds) < 4:
            raise ValueError("Minimum 4 seeds are required.")

        from ._spherical_voronoi import _check_if_in_bounds, _get_duplicated_seeds_ids

        duplicated_seeds_ids = _get_duplicated_seeds_ids(self.seeds, self.region_ids)
        if duplicated_seeds_ids:
            raise ValueError(f"Duplicate seeds present: {duplicated_seeds_ids}.")

        if not _check_if_in_bounds(self.seeds):
            raise ValueError("Seeds outside Earth WGS84 bounding box.")

    def transform(self, gdf: Optional[gpd.GeoDataFrame] = None) -> gpd.GeoDataFrame:
        """
        Regionalize a given GeoDataFrame.

        Returns a list of disjointed regions consisting of Thiessen cells generated
        using a Voronoi diagram on the sphere.

        Args:
            gdf (Optional[gpd.GeoDataFrame], optional): GeoDataFrame to be regionalized.
                Will use this list of geometries to crop resulting regions. If None, a boundary box
                with bounds (-180, -90, 180, 90) is used to return regions covering whole Earth.
                Defaults to None.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with the regionalized data cropped using input
                GeoDataFrame.

        Raises:
            ValueError: If provided geodataframe has no crs defined.
            ValueError: If seeds are laying on a single arc.
        """
        from ._spherical_voronoi import generate_voronoi_regions

        if gdf is None:
            gdf = gpd.GeoDataFrame(
                {GEOMETRY_COLUMN: [box(minx=-180, maxx=180, miny=-90, maxy=90)]}, crs=WGS84_CRS
            )

        gdf_wgs84 = gdf.to_crs(crs=WGS84_CRS)
        generated_regions = generate_voronoi_regions(
            seeds=self.seeds,
            max_meters_between_points=self.max_meters_between_points,
            num_of_multiprocessing_workers=self.num_of_multiprocessing_workers,
            multiprocessing_activation_threshold=self.multiprocessing_activation_threshold,
        )
        regions_gdf = gpd.GeoDataFrame(
            data={GEOMETRY_COLUMN: generated_regions}, index=self.region_ids, crs=WGS84_CRS
        )
        regions_gdf.index.rename(REGIONS_INDEX, inplace=True)
        clipped_regions_gdf = regions_gdf.clip(mask=gdf_wgs84, keep_geom_type=False)
        return clipped_regions_gdf



"""
Embedding model for Highway2Vec.

This module contains the embedding model from the `highway2vec` paper [1].

References:
    1. https://doi.org/10.1145/3557918.3565865
"""

from typing import TYPE_CHECKING

from srai._optional import import_optional_dependencies
from srai.embedders import Model

if TYPE_CHECKING:  # pragma: no cover
    import torch


class Highway2VecModel(Model):
    """Autoencoder based embedding model for highway2vec."""

    def __init__(self, n_features: int, n_hidden: int = 64, n_embed: int = 30, lr: float = 1e-3):
        """
        Init Highway2VecModel.

        Args:
            n_features (int): Number of features.
            n_hidden (int, optional): Number of hidden units. Defaults to 64.
            n_embed (int, optional): Embedding size. Defaults to 30.
            lr (float, optional): Learning rate. Defaults to 1e-3.
        """
        import_optional_dependencies(
            dependency_group="torch", modules=["torch", "pytorch_lightning"]
        )
        from torch import nn

        super().__init__()

        self.save_hyperparameters()

        self.encoder = nn.Sequential(
            nn.Linear(n_features, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, n_embed),
        )
        self.decoder = nn.Sequential(
            nn.Linear(n_embed, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, n_features),
        )
        self.n_features = n_features
        self.n_hidden = n_hidden
        self.n_embed = n_embed
        self.lr = lr

    def forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Forward pass.

        Args:
            x (torch.Tensor): Input tensor.
        """
        z: torch.Tensor = self.encoder(x)
        return z

    def training_step(self, batch: "torch.Tensor", batch_idx: int) -> "torch.Tensor":
        """
        Training step.

        Args:
            batch (torch.Tensor): Batch.
            batch_idx (int): Batch index.
        """
        return self._common_step(batch, batch_idx, "train")

    def _common_step(self, batch: "torch.Tensor", batch_idx: int, stage: str) -> "torch.Tensor":
        """
        Perform common step.

        Args:
            batch (torch.Tensor): Batch.
            batch_idx (int): Batch index.
            stage (str): Name of the stage - e.g. train, valid, test.
        """
        import torch.nn.functional as F

        x = batch
        z = self.encoder(x)
        x_hat = self.decoder(z)
        loss = F.mse_loss(x_hat, x)

        self.log(f"{stage}_loss", loss, on_epoch=True, on_step=True, prog_bar=True)

        return loss

    def configure_optimizers(self) -> "torch.optim.Optimizer":
        """Configure optimizer."""
        import torch

        optimizer = torch.optim.Adam(self.parameters(), lr=self.lr)
        return optimizer



"""
Adjacency neighbourhood.

This module contains the AdjacencyNeighbourhood class, that allows to get the neighbours of any
region based on its borders.
"""

from collections.abc import Hashable
from typing import Optional

import geopandas as gpd

from srai.constants import GEOMETRY_COLUMN
from srai.neighbourhoods import Neighbourhood


class AdjacencyNeighbourhood(Neighbourhood[Hashable]):
    """
    Adjacency Neighbourhood.

    This class allows to get the neighbours of any region based on common border. Additionally, a
    lookup table is implemented to accelerate repeated queries.

    By default, a lookup table will be populated lazily based on queries. A dedicated function
    `generate_neighbourhoods` allows for precalculation of all the neighbourhoods at once.
    """

    def __init__(self, regions_gdf: gpd.GeoDataFrame, include_center: bool = False) -> None:
        """
        Init AdjacencyNeighbourhood.

        Args:
            regions_gdf (gpd.GeoDataFrame): regions for which a neighbourhood will be calculated.
            include_center (bool): Whether to include the region itself in the neighbours.
            This is the default value used for all the methods of the class,
            unless overridden in the function call.

        Raises:
            ValueError: If regions_gdf doesn't have geometry column.
        """
        super().__init__(include_center)
        if GEOMETRY_COLUMN not in regions_gdf.columns:
            raise ValueError("Regions must have a geometry column.")
        self.regions_gdf = regions_gdf
        self.lookup: dict[Hashable, set[Hashable]] = {}

    def generate_neighbourhoods(self) -> None:
        """Generate the lookup table for all regions."""
        for region_id in self.regions_gdf.index:
            if region_id not in self.lookup:
                self.lookup[region_id] = self._get_adjacent_neighbours(region_id)

    def get_neighbours(
        self, index: Hashable, include_center: Optional[bool] = None
    ) -> set[Hashable]:
        """
        Get the direct neighbours of any region using its index.

        Args:
            index (Hashable): Unique identifier of the region.
            include_center (Optional[bool]): Whether to include the region itself in the neighbours.
            If None, the value set in __init__ is used. Defaults to None.

        Returns:
            Set[Hashable]: Indexes of the neighbours.
        """
        if self._index_incorrect(index):
            return set()

        if index not in self.lookup:
            self.lookup[index] = self._get_adjacent_neighbours(index)

        neighbours = self.lookup[index]
        neighbours = self._handle_center(
            index, 1, neighbours, at_distance=False, include_center_override=include_center
        )
        return neighbours

    def _get_adjacent_neighbours(self, index: Hashable) -> set[Hashable]:
        """
        Get the direct neighbours of a region using `touches` [1] operator from the Shapely library.

        Args:
            index (Hashable): Unique identifier of the region.

        Returns:
            Set[Hashable]: Indexes of the neighbours.

        References:
            1. https://shapely.readthedocs.io/en/stable/reference/shapely.touches.html
        """
        current_region = self.regions_gdf.loc[index]
        neighbours = self.regions_gdf[
            self.regions_gdf.geometry.touches(current_region[GEOMETRY_COLUMN])
        ]
        return set(neighbours.index)

    def _index_incorrect(self, index: Hashable) -> bool:
        return index not in self.regions_gdf.index



"""Tests for OSMPbfLoader."""

from pathlib import Path
from typing import Union
from unittest import TestCase

import geopandas as gpd
import pytest
from shapely.geometry import Polygon
from shapely.geometry.base import BaseGeometry

from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS
from srai.loaders.osm_loaders import OSMPbfLoader
from srai.loaders.osm_loaders.filters import (
    GEOFABRIK_LAYERS,
    HEX2VEC_FILTER,
    GroupedOsmTagsFilter,
    OsmTagsFilter,
)

ut = TestCase()


@pytest.mark.parametrize("explode_tags", [True, False])  # type: ignore
@pytest.mark.parametrize("keep_all_tags", [True, False])  # type: ignore
def test_pbf_to_geoparquet_parsing(explode_tags: bool, keep_all_tags: bool):
    """Test if pbf to geoparquet conversion works."""
    pbf_file = Path(__file__).parent / "test_files" / "monaco.osm.pbf"
    OSMPbfLoader(pbf_file=pbf_file).load_to_geoparquet(
        area=Polygon(
            [
                (7.416769421059001, 43.7346112362936),
                (7.416769421059001, 43.730681304758946),
                (7.4218262821731, 43.730681304758946),
                (7.4218262821731, 43.7346112362936),
            ]
        ),
        tags=GEOFABRIK_LAYERS,
        ignore_cache=True,
        explode_tags=explode_tags,
        keep_all_tags=keep_all_tags,
    )


@pytest.mark.parametrize(  # type: ignore
    "test_geometries,pbf_file,query,pbf_source,expected_result_length,expected_features_columns_length,expected_features_columns_names",
    [
        (
            [
                Polygon(
                    [
                        (7.416769421059001, 43.7346112362936),
                        (7.416769421059001, 43.730681304758946),
                        (7.4218262821731, 43.730681304758946),
                        (7.4218262821731, 43.7346112362936),
                    ]
                )
            ],
            Path(__file__).parent / "test_files" / "monaco.osm.pbf",
            HEX2VEC_FILTER,
            "geofabrik",
            403,
            12,
            [
                "amenity",
                "building",
                "healthcare",
                "historic",
                "landuse",
                "leisure",
                "natural",
                "office",
                "shop",
                "sport",
                "tourism",
                "water",
            ],
        ),
        (
            [
                Polygon(
                    [
                        (7.416769421059001, 43.7346112362936),
                        (7.416769421059001, 43.730681304758946),
                        (7.4218262821731, 43.730681304758946),
                        (7.4218262821731, 43.7346112362936),
                    ]
                )
            ],
            Path(__file__).parent / "test_files" / "monaco.osm.pbf",
            GEOFABRIK_LAYERS,
            "geofabrik",
            958,
            23,
            [
                "accommodation",
                "buildings",
                "catering",
                "education",
                "fuel_parking",
                "health",
                "highway_links",
                "landuse",
                "leisure",
                "major_roads",
                "minor_roads",
                "miscpoi",
                "money",
                "natural",
                "paths_unsuitable_for_cars",
                "public",
                "shopping",
                "tourism",
                "traffic",
                "transport",
                "very_small_roads",
                "water",
                "water_traffic",
            ],
        ),
    ],
)
def test_osm_pbf_loader(
    test_geometries: list[BaseGeometry],
    pbf_file: Path,
    query: Union[OsmTagsFilter, GroupedOsmTagsFilter],
    pbf_source: str,
    expected_result_length: int,
    expected_features_columns_length: int,
    expected_features_columns_names: list[str],
):
    """Test `OSMPbfLoader.load()`."""
    download_directory = Path(__file__).parent / "test_files"
    area = gpd.GeoDataFrame(
        geometry=test_geometries,
        index=gpd.pd.Index(name=REGIONS_INDEX, data=list(range(len(test_geometries)))),
        crs=WGS84_CRS,
    )

    loader = OSMPbfLoader(
        pbf_file=pbf_file,
        download_directory=download_directory,
        download_source=pbf_source,
    )
    result = loader.load(area, tags=query, ignore_cache=True)

    assert (
        len(result) == expected_result_length
    ), f"Mismatched result length ({len(result)}, {expected_result_length})"
    assert (
        len(result.columns) == expected_features_columns_length + 1
    ), f"Mismatched columns length ({len(result.columns)}, {expected_features_columns_length + 1})"
    ut.assertCountEqual(
        result.columns,
        expected_features_columns_names + [GEOMETRY_COLUMN],
        "Mismatched columns names.",
    )
